#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2020/05/30 20:37

from pyxllib.prog.pupil import check_install_package

check_install_package('filetype')

from typing import Callable, Any
import io
import json
import os
import pathlib
import pickle
import re
import shutil
import subprocess
import tempfile
import ujson
from collections import defaultdict, Counter
import math
from itertools import islice
import datetime

# import chardet
import charset_normalizer
import qiniu
import requests
import yaml
import humanfriendly
from more_itertools import chunked
import filetype
from tqdm import tqdm

from pyxllib.prog.newbie import round_int, human_readable_size
from pyxllib.prog.pupil import is_url, is_file, DictTool
from pyxllib.algo.pupil import Groups
from pyxllib.file.pupil import struct_unpack, gen_file_filter


def __1_judge():
    pass


def is_url_connect(url, timeout=5):
    try:
        _ = requests.head(url, timeout=timeout)
        return True
    except requests.ConnectionError:
        pass
    return False


def __2_qiniu():
    pass


class GetEtag:
    """ 七牛原有etag功能基础上做封装 """

    @classmethod
    def from_bytes(cls, _bytes):
        return qiniu.utils.etag_stream(io.BytesIO(_bytes))

    @classmethod
    def from_text(cls, text):
        _bytes = text.encode('utf8')
        return qiniu.utils.etag_stream(io.BytesIO(_bytes))

    @classmethod
    def from_file(cls, file):
        return qiniu.etag(file)

    @classmethod
    def from_url(cls, url):
        return cls(requests.get(url).content)


# @deprecated.deprecated
def get_etag(arg):
    """ 七牛原有etag功能基础上做封装

    :param arg: 支持bytes二进制、文件、url地址

    只跟文件内容有关，跟文件创建、修改日期没关系
    如果读取文件后再处理etag，要尤其小心 '\r\n' 的问题！
    文件里如果是\r\n，我的File.read会变成\n，所以按文件取etag和read的内容算etag会不一样。
    """
    if isinstance(arg, bytes):  # 二进制数据
        return qiniu.utils.etag_stream(io.BytesIO(arg))
    elif is_file(arg):  # 输入是一个文件
        return qiniu.etag(arg)
    elif is_url(arg):  # 输入是一个网页上的数据源
        return get_etag(requests.get(arg).content)
    elif isinstance(arg, str):  # 明文字符串转二进制
        return get_etag(arg.encode('utf8'))
    else:
        raise TypeError('不识别的数据类型')


def is_etag(s):
    """ 字母、数字和-、_共64种字符构成的长度28的字符串 """
    return re.match(r'[a-zA-Z0-9\-_]{28}$', s)


def test_etag():
    print(get_etag(r'\chematom{+8}{2}{8}{}'))
    # Fjnu-ZXyDxrqLoZmNJ2Kj8FcZGR-

    print(get_etag(__file__))
    # 每次代码改了这段输出都是不一样的


def test_etag2():
    """ 字符串值和写到文件判断的etag，是一样的

    """
    s = 'code4101'
    print(get_etag(s))
    # FkAD2McB6ugxTiniE8ebhlNHdHh9

    f = File('1.tex', tempfile.gettempdir()).write(s, if_exists='replace').to_str()
    print(get_etag(f))
    # FkAD2McB6ugxTiniE8ebhlNHdHh9


def __3_chardet():
    pass


def get_encoding(data,
                 cp_isolation=('utf_8', 'gbk', 'gb18030', 'utf_16'),
                 preemptive_behaviour=True,
                 explain=False):
    """ 从字节串中检测编码类型

    :param bytes data: 要检测的字节串
    :param List[str] cp_isolation: 指定要检测的字符集编码类型列表，只检测该列表中指定的编码类型，而不是所有可能的编码类型，默认为 None
        注意charset_normalizer是无法识别utf-8-sig的情况的。但要获得解析后的文本内容，其实也不用通过get_encoding来中转。
    :param bool preemptive_behaviour: 指定预处理行为，如果设置为 True，将在检测之前对数据进行预处理，例如去除 BOM、转换大小写等操作，默认为 True
    :param bool explain: 指定是否打印出检测过程的详细信息，如果设置为 True，将打印出每个 chunk 的检测结果和置信度，默认为 False
    :return str: 检测到的编码类型，返回字符串表示
    """
    result = charset_normalizer.from_bytes(data,
                                           cp_isolation=cp_isolation,
                                           preemptive_behaviour=preemptive_behaviour,
                                           explain=explain)
    best_match = result.best()
    if best_match:
        return best_match.encoding
    else:  # 注意，这个实现是有可能会找不到编码的，此时默认返回None
        return


def __4_file():
    """
    路径、文件、目录相关操作功能

    主要是为了提供readfile、wrritefile函数
    与普通的读写文件相比，有以下优点：
    1、智能识别pkl等特殊格式文件的处理
    2、智能处理编码
    3、目录不存在自动创建
    4、自动备份旧文件，而不是强制覆盖写入

    其他相关文件处理组件：isfile、get_encoding、ensure_folders
    """


class PathBase:
    """ File和Dir共有的操作逻辑功能 """
    __slots__ = ('_path',)

    @classmethod
    def abspath(cls, path=None, root=None, *, suffix=None, resolve=True) -> pathlib.Path:
        r""" 根据各种不同的组合参数信息，推导出具体的路径位置

        :param path: 主要的参数，如果后面的参数有矛盾，以path为最高参考标准
            ...，可以输入Ellipsis对象，效果跟None是不一样的，意思是显式地指明要生成一个随机名称的文件
            三个参数全空时，则返回当前目录
        :param suffix:
            以 '.' 指明的扩展名，会强制替换
            否则，只作为参考扩展名，只在原有path没有指明的时候才添加
        :param root: 未输入的时候，则为当前工作目录

        >>> os.chdir("C:/Users")

        # 未输入任何参数，则返回当前工作目录
        >>> PathBase.abspath()
        WindowsPath('C:/Users')

        # 基本的指定功能
        >>> PathBase.abspath('a')
        WindowsPath('C:/Users/a')

        # 额外指定父目录
        >>> PathBase.abspath('a/b', 'D:')
        WindowsPath('D:/a/b')

        # 设置扩展名
        >>> PathBase.abspath('a', suffix='.txt')
        WindowsPath('C:/Users/a.txt')

        # 扩展名的高级用法
        >>> PathBase.abspath('F:/work/a.txt', suffix='py')  # 参考后缀不修改
        WindowsPath('F:/work/a.txt')
        >>> PathBase.abspath('F:/work/a.txt', suffix='.py')  # 强制后缀会修改
        WindowsPath('F:/work/a.py')
        >>> PathBase.abspath('F:/work/a.txt', suffix='')  # 删除后缀
        WindowsPath('F:/work/a')

        # 在临时目录下，新建一个.tex的随机名称文件
        >> PathBase.abspath(..., tempfile.gettempdir(), suffix='.tex')
        WindowsPath('D:/Temp/tmp_sey0yeg.tex')
        """
        # 1 判断参考目录
        if root is None:
            root = os.getcwd()
        else:
            root = str(root)

        # 2 判断主体文件名 path
        if path is None:
            return pathlib.Path(root)
        elif path is Ellipsis:
            path = tempfile.mktemp(dir=root)
        else:
            path = os.path.join(root, str(path))

        # 3 补充suffix
        if suffix is not None:
            # 判断后缀
            li = os.path.splitext(path)
            if suffix == '' or suffix[0] == '.':
                path = li[0] + suffix
            elif li[1] == '':
                path = li[0] + '.' + suffix

        if resolve:
            return pathlib.Path(path).resolve()
        else:
            return pathlib.Path(path)

    def __bool__(self):
        r""" 判断文件、文件夹是否存在

        重置WindowsPath的bool逻辑，返回值变成存在True，不存在为False

        >>> bool(File('C:/Windows/System32/cmd.exe'))
        True
        >>> bool(File('C:/Windows/System32/cmdcmd.exe'))
        False
        """
        return self._path.exists()

    def __repr__(self):
        s = self._path.__repr__()
        if s.startswith('WindowsPath'):
            s = self.__class__.__name__ + s[11:]
        elif s.startswith('PosixPath'):
            s = self.__class__.__name__ + s[9:]
        return s

    def __str__(self):
        return str(self._path).replace('\\', '/')

    def to_str(self):
        return self.__str__()

    def resolve(self):
        return self._path.resolve()

    @property
    def drive(self) -> str:
        return self._path.drive

    @drive.setter
    def drive(self, value):
        """修改磁盘位置"""
        raise NotImplementedError

    @property
    def name(self) -> str:
        r"""
        >>> File('D:/pycode/a.txt').name
        'a.txt'
        >>> File('D:/pycode/code4101py').name
        'code4101py'
        >>> File('D:/pycode/.gitignore').name
        '.gitignore'
        """
        return self._path.name

    @name.setter
    def name(self, value):
        raise NotImplementedError

    @property
    def parent(self):
        r"""
        >>> File('D:/pycode/code4101py').parent
        WindowsPath('D:/pycode')
        """
        return self._path.parent

    @property
    def dirname(self) -> str:
        r"""
        >>> File('D:/pycode/code4101py').dirname
        'D:\\pycode'
        >>> File(r'D:\toweb\a').dirname
        'D:\\toweb'
        """
        return str(self.parent)

    @property
    def parts(self) -> tuple:
        r"""
        >>> File('D:/pycode/code4101py').parts
        ('D:\\', 'pycode', 'code4101py')
        """
        return self._path.parts

    @property
    def backup_time(self):
        r""" 返回文件的备份时间戳，如果并不是备份文件，则返回空字符串

        备份文件都遵循特定的命名规范
        如果是文件，是：'chePre 171020-153959.tex'
        如果是目录，是：'figs 171020-153959'
        通过后缀分析，可以判断这是不是一个备份文件

        >>> File('chePre 171020-153959.tex').backup_time
        '171020-153959'
        >>> File('figs 171020-153959').backup_time
        '171020-153959'
        >>> File('figs 171020').backup_time
        ''
        """
        name = self.stem
        if len(name) < 14:
            return ''
        g = re.match(r'(\d{6}-\d{6})', name[-13:])
        return g.group(1) if g else ''

    @property
    def mtime(self):
        r""" 文件的最近修改时间

        >> PathBase(r"C:\pycode\code4101py").mtime
        datetime.datetime(2021, 9, 6, 17, 28, 18, 960713)
        """
        from datetime import datetime
        return datetime.fromtimestamp(os.stat(str(self)).st_mtime)

    @property
    def ctime(self):
        r""" 文件的创建时间

        >> Path(r"C:\pycode\code4101py").ctime
        2018-05-25 10:46:37
        """
        from datetime import datetime
        # 注意：st_ctime是平台相关的值，在windows是创建时间，但在Unix是metadate最近修改时间
        return datetime.fromtimestamp(os.stat(str(self)).st_ctime)

    def relpath(self, ref_dir) -> str:
        r""" 当前路径，相对于ref_dir的路径位置

        >>> File('C:/a/b/c.txt').relpath('C:/a/')
        'b/c.txt'
        >>> File('C:/a/b\\c.txt').relpath('C:\\a/')
        'b/c.txt'

        >> File('C:/a/b/c.txt').relpath('D:/')  # ValueError
        """
        return os.path.relpath(str(self), str(ref_dir)).replace('\\', '/')

    def exist_preprcs(self, if_exists=None):
        """ 这个实际上是在做copy等操作前，如果目标文件已存在，需要预先删除等的预处理
        并返回判断，是否需要执行下一步操作

        有时候情况比较复杂，process无法满足需求时，可以用exist_preprcs这个底层函数协助

        :param if_exists:
            None: 不做任何处理，直接运行，依赖于功能本身是否有覆盖写入机制
            'error': 如果要替换的目标文件已经存在，则报错
            'replace': 把存在的文件先删除
                本来是叫'delete'更准确的，但是考虑用户理解，
                    一般都是用在文件替换场合，叫成'delete'会非常怪异，带来不必要的困扰、误解
                所以还是决定叫'replace'
            'skip': 不执行后续功能
            'backup': 先做备份  （对原文件先做一个备份）
        """
        need_run = True
        if self.exists():
            if if_exists is None:
                return need_run
            elif if_exists == 'error':
                raise FileExistsError(f'目标文件已存在： {self}')
            elif if_exists == 'replace':
                self.delete()
            elif if_exists == 'skip':
                need_run = False
            elif if_exists == 'backup':
                self.backup(move=True)
            else:
                raise ValueError(f'{if_exists}')
        return need_run

    def copy(self, *args, **kwargs):
        raise NotImplementedError

    def delete(self):
        raise NotImplementedError

    def absdst(self, dst):
        raise NotImplementedError

    def backup(self, tail=None, if_exists='replace', move=False):
        r""" 对文件末尾添加时间戳备份，也可以使用自定义标记tail

        :param tail: 自定义添加后缀
            tail为None时，默认添加特定格式的时间戳
        :param if_exists: 备份的目标文件名存在时的处理方案
            这个概率非常小，真遇到，先把已存在的删掉，重新写入一个是可以接受的
        :param move: 是否删除原始文件

        # TODO：有个小bug，如果在不同时间实际都是相同一个文件，也会被不断反复备份
        #    如果想解决这个，就要读取目录下最近的备份文件对比内容了
        """
        # 1 判断自身文件是否存在
        if not self:
            return None

        # 2 计算出新名称
        if not tail:
            tail = self.mtime.strftime(' %y%m%d-%H%M%S')  # 时间戳
        name, ext = os.path.splitext(str(self))
        dst = name + tail + ext

        # 3 备份就是特殊的copy操作
        if move:
            return self.move(dst, if_exists)
        else:
            return self.copy(dst, if_exists)

    def move(self, dst, if_exists=None):
        """ 移动文件
        """
        if self:
            return self.process(dst, shutil.move, if_exists)

    def process(self, dst, func, if_exists=None):
        r""" copy或move的本质底层实现

        :param dst: 目标路径对象
        :param func: 传入arg1和arg2参数，可以自定义
            默认分别是self和dst的字符串
        :return : 返回dst
        """
        # 1 判断目标是否已存在，进行不同的指定规则处理
        dst_ = self.absdst(dst)

        # 2 执行特定功能
        if self == dst_:
            # 如果文件是自身的话，并不算exists，可以直接run，不用exist_preprcs
            dst_.ensure_parent()
            func(str(self), str(dst_))
            dst_._path = dst_._path.resolve()  # 但是需要重新解析一次dst_._path，避免可能有重命名等大小写变化
        elif dst_.exist_preprcs(if_exists):
            dst_.ensure_parent()
            func(str(self), str(dst_))

        return dst_

    def explorer(self, proc='explorer'):
        """ 使用windows的explorer命令打开文件

        还有个类似的万能打开命令 start

        :param proc: 可以自定义要执行的主程序
        """
        subprocess.run([proc, str(self)])

    def ensure_parent(self):
        r""" 确保父目录存在
        """
        p = self.parent
        if not p.exists():
            os.makedirs(str(p))

    def __getattr__(self, item):
        return getattr(self._path, item)

    def start(self):
        os.startfile(str(self))


class File(PathBase):
    r""" 通用文件处理类，大部分基础功能是从pathlib.Path衍生过来的

    document: https://www.yuque.com/xlpr/python/pyxllib.debug.path
    """
    __slots__ = ('_path',)

    # 一、基础功能

    def __init__(self, path, root=None, *, suffix=None, check=True):
        r""" 初始化参数含义详见 PathBase.abspath 函数解释

        :param path: 只传入一个File、pathlib.Path对象，可以提高初始化速度，不会进行多余的解析判断
            如果要安全保守一点，可以传入str类型的path
        :param root: 父目录
        :param check: 在某些特殊场合、内部开发中，可以确保传参一定不会出错，在上游已经有了严谨的检查
            此时可以设置check=False，提高初始化速度

        注意！如果使用了符号链接（软链接），则路径是会解析转向实际位置的！例如
        >> File('D:/pycode/code4101py')
        File('D:/slns/pycode/code4101py')
        """
        self._path = None

        # 1 快速初始化
        if root is None and suffix is None:
            if isinstance(path, File):
                self._path = path._path
                return  # 直接完成初始化过程
            elif isinstance(path, pathlib.Path):
                self._path = path

        # 2 普通初始化
        if self._path is None:
            self._path = self.abspath(path, root, suffix=suffix)

        # 3 检查
        if check:
            if not self._path:
                raise ValueError(f'无效路径 {self._path}')
            elif self._path.is_dir():
                raise ValueError(f'不能用目录初始化一个File对象 {self._path}')

    @classmethod
    def safe_init(cls, path, root=None, *, suffix=None):
        """ 如果失败不raise，而是返回None的初始化方式 """
        try:
            f = File(path, root, suffix=suffix)
            f._path.is_file()  # 有些问题上一步不一定测的出来，要再补一个测试
            return f
        except (ValueError, TypeError, OSError, PermissionError):
            # ValueError：文件名过长，代表输入很可能是一段文本，根本不是路径
            # TypeError：不是str等正常的参数
            # OSError：非法路径名，例如有 *? 等
            # PermissionError: linux上访问无权限、不存在的路径
            return None

    # 二、获取、修改路径中部分值的功能

    def with_dirname(self, value):
        return File(self.name, value)

    @property
    def stem(self) -> str:
        r"""
        >>> File('D:/pycode/code4101py/ckz.py').stem
        'ckz'
        >>> File('D:/pycode/.gitignore').stem  # os.path.splitext也是这种算法
        '.gitignore'
        >>> File('D:/pycode/.123.45.6').stem
        '.123.45'
        """
        return self._path.stem

    def with_stem(self, stem):
        """
        注意不能用@stem.setter来做
            如果setter要rename，那if_exists参数怎么控制？
            如果setter不要rename，那和用with_stem实现有什么区别？
        """
        return File(stem, self.parent, suffix=self.suffix)

    def with_name(self, name):
        return File(name, self.parent)

    @property
    def suffix(self) -> str:
        r"""
        >>> File('D:/pycode/code4101py/ckz.py').suffix
        '.py'
        >>> File('D:/pycode/code4101py').suffix
        ''
        >>> File('D:/pycode/code4101py/ckz.').suffix
        ''
        >>> File('D:/pycode/code4101py/ckz.123.456').suffix
        '.456'
        >>> File('D:/pycode/code4101py/ckz.123..456').suffix
        '.456'
        >>> File('D:/pycode/.gitignore').suffix
        ''
        """
        return self._path.suffix

    def with_suffix(self, suffix):
        r""" 指向同目录下后缀为suffix的文件

        >>> File('a.txt').with_suffix('.py').name  # 强制替换
        'a.py'
        >>> File('a.txt').with_suffix('py').name  # 参考替换
        'a.txt'
        >>> File('a.txt').with_suffix('').name  # 删除
        'a'
        """
        return File(self.abspath(self._path, suffix=suffix))

    # 三、获取文件相关属性值功能

    # @property
    # def encoding(self):
    #     """ 文件的编码
    #
    #     非文件、不存在时返回 None
    #     """
    #     if self:
    #         return get_encoding(str(self))

    @property
    def size(self) -> int:
        """ 计算文件大小
        """
        if self.exists():
            total_size = os.path.getsize(str(self))
        else:
            total_size = 0
        return total_size

    def size2(self) -> int:
        """ 220102周日17:23

        size有点bug，临时写个函数接口
        这个bug有点莫名其妙，搞不定
        """
        total_size = os.path.getsize(str(self))
        return total_size

    # 四、文件操作功能

    def absdst(self, dst):
        """ 在copy、move等中，给了个"模糊"的目标位置dst，智能推导出实际file、dir绝对路径
        """
        from pyxllib.file.specialist.dirlib import Dir
        dst_ = self.abspath(dst)
        if isinstance(dst, Dir) or (isinstance(dst, str) and dst[-1] in ('\\', '/')) or dst_.is_dir():
            dst_ = File(self.name, dst_)
        else:
            dst_ = File(dst_)
        return dst_

    def copy(self, dst, if_exists=None):
        """ 复制文件
        """
        return self.process(dst, shutil.copy2, if_exists)

    def rename(self, dst, if_exists=None):
        r""" 文件重命名，或者也可以理解成文件移动
        该接口和move的核心区别：move的dst是相对工作目录，而rename则是相对self.parent路径
        """
        # rename是move的一种特殊情况
        return self.move(File(dst, self.parent), if_exists)

    def delete(self):
        r""" 删除自身文件
        """
        if self.is_file():
            os.remove(str(self))

    # 五、其他综合性功能

    def read(self, *, encoding=None, mode=None):
        """ 读取文件

        :param encoding: 文件编码
            默认None，则在需要使用encoding参数的场合，会使用self.encoding自动判断编码
        :param mode: 读取模式（例如 '.json'），默认从扩展名识别，也可以强制指定
            'b': 特殊标记，表示按二进制读取文件内容
        :return:
        """
        if self:  # 如果存在这样的文件，那就读取文件内容
            # 获得文件扩展名，并统一转成小写
            name, suffix = str(self), self.suffix
            if not mode: mode = suffix
            mode = mode.lower()
            if mode == 'bytes':
                with open(name, 'rb') as f:
                    return f.read()
            elif mode == '.pkl':  # pickle库
                with open(name, 'rb') as f:
                    return pickle.load(f)
            elif mode == '.json':
                # 先读成字符串，再解析，会比rb鲁棒性更强，能自动过滤掉开头可能非正文特殊标记的字节
                with open(name, 'rb') as f:
                    bstr = f.read()
                    if not encoding: encoding = get_encoding(bstr)
                try:
                    return ujson.loads(bstr.decode(encoding=encoding))
                except ValueError:  # ujson会有些不太标准的情况处理不了
                    return json.loads(bstr.decode(encoding=encoding))
            elif mode == '.yaml':
                with open(name, 'r', encoding=encoding) as f:
                    return yaml.safe_load(f.read())
            elif mode in ('.jpg', '.jpeg', '.png', '.bmp', 'b'):
                # 二进制读取
                with open(name, 'rb') as fp:
                    return fp.read()
            else:
                with open(name, 'rb') as f:
                    bstr = f.read()
                if not encoding:
                    encoding = get_encoding(bstr)
                    if not encoding:
                        raise ValueError(f'{self} 自动识别编码失败，请手动指定文件编码')
                s = bstr.decode(encoding=encoding, errors='ignore')
                if '\r' in s: s = s.replace('\r\n', '\n')  # 如果用\r\n作为换行符会有一些意外不好处理
                return s
        else:  # 非文件对象
            raise FileNotFoundError(f'{self} 文件不存在，无法读取。')

    def write(self, ob, *, encoding='utf8', if_exists=None, mode=None, **kwargs):
        """ 保存为文件

        :param ob: 写入的内容
            如果要写txt文本文件且ob不是文本对象，只会进行简单的字符串化
        :param encoding: 强制写入的编码
            如果原文件存在且有编码，则使用原文件的编码
            如果没有，则默认使用utf8
            当然，其实有些格式是用不到编码信息的~~例如pkl文件
        :param if_exists: 如果文件已存在，要进行的操作
        :param mode: 写入模式（例如 '.json'），默认从扩展名识别，也可以强制指定
        :param kwargs:
            写入json格式的时候
                ensure_ascii: json.dump默认是True，但是我这里默认值改成了False
                    改成False可以支持在json直接显示中文明文
                indent: json.dump是None，我这里默认值遵循json.dump
                    我原来是2，让文件结构更清晰、更加易读
        :return: 返回写入的文件名，这个主要是在写临时文件时有用
        """

        # # 将ob写入文件path
        # def get_enc():
        #     # 编码在需要的时候才获取分析，减少不必要的运算开销
        #     # 所以封装一个函数接口，需要的时候再计算
        #     if encoding is None:
        #         # return self.encoding or 'utf8'
        #     return encoding

        if self.exist_preprcs(if_exists):
            self.ensure_parent()
            name, suffix = str(self), self.suffix
            if not mode: mode = suffix
            mode = mode.lower()
            if mode == '.pkl':
                with open(name, 'wb') as f:
                    pickle.dump(ob, f)
            elif mode == '.json':
                with open(name, 'w', encoding=encoding) as f:
                    DictTool.ior(kwargs, {'ensure_ascii': False})
                    json.dump(ob, f, **kwargs)
            elif mode == '.yaml':
                with open(name, 'w', encoding=encoding) as f:
                    yaml.dump(ob, f)
            elif isinstance(ob, bytes):
                with open(name, 'wb') as f:
                    f.write(ob)
            else:  # 其他类型认为是文本类型
                with open(name, 'w', errors='ignore', encoding=encoding) as f:
                    f.write(str(ob))

        return self

    def unpack(self, dst_dir=None):
        """ 解压缩

        :param dst_dir: 如果没有输入，默认会套一层压缩文件原名的目录里

        TODO 本来是想，如果没有传dst_dir，则类似Bandizip的自动解压机制，解压到当前文件夹
            但是不太好实现
        """
        p = str(self)
        if not dst_dir:
            dst_dir = os.path.splitext(p)[0]
        shutil.unpack_archive(p, str(dst_dir))

    def __eq__(self, other):
        return str(self) == str(other)

    def exists(self):
        return self._path.exists()


def make_filter():
    """ 从filesmatch """


class XlPath(type(pathlib.Path())):

    @classmethod
    def desktop(cls):
        if os.getenv('Desktop', None):  # 如果修改了win10默认的桌面路径，需要在环境变量添加一个正确的Desktop路径值
            desktop = os.environ['Desktop']
        else:
            desktop = os.path.join(pathlib.Path.home(), 'Desktop')  # 这个不一定准，桌面是有可能被移到D盘等的
        return cls(desktop)

    @classmethod
    def userdir(cls):
        from os.path import expanduser
        return cls(expanduser("~"))

    @classmethod
    def tempdir(cls):
        return cls(tempfile.gettempdir())

    @classmethod
    def create_tempdir_path(cls, dir=None):
        if dir is None:
            dir = tempfile.gettempdir()
        dst = cls(tempfile.mktemp(dir=dir))
        return dst

    @classmethod
    def create_tempfile_path(cls, suffix="", dir=None):
        if dir is None:
            dir = tempfile.gettempdir()
        return cls(tempfile.mktemp(suffix=suffix, dir=dir))

    tempfile = create_tempfile_path

    @classmethod
    def init(cls, path, root=None, *, suffix=None):
        """ 仿照原来File的初始化接口形式 """
        p = XlPath(path)
        if root:
            p = XlPath(root) / p

        if suffix:
            if suffix[0] == '.':  # 使用.时强制修改后缀
                p = p.with_suffix(suffix)
            elif not p.suffix:
                p = p.with_suffix('.' + suffix)

        return p

    @classmethod
    def safe_init(cls, arg_in):
        """ 输入任意类型的_in，会用比较安全的机制，判断其是否为一个有效的路径格式并初始化
        初始化失败则返回None
        """
        try:
            p = XlPath(str(arg_in))
            p.is_file()  # 有些问题上一步不一定测的出来，要再补一个测试。具体存不存在是不是文件并不重要，而是使用这个能检查出问题。
            return p
        except (ValueError, TypeError, OSError, PermissionError):
            # ValueError：文件名过长，代表输入很可能是一段文本，根本不是路径
            # TypeError：不是str等正常的参数
            # OSError：非法路径名，例如有 *? 等
            # PermissionError: linux上访问无权限、不存在的路径
            return None

    def start(self, *args, **kwargs):
        """ 使用关联的程序打开p，类似于双击的效果

        这就像在 Windows 资源管理器中双击文件、文件夹
        """
        os.startfile(self, *args, **kwargs)

    def exists_type(self):
        """
        不存在返回0，文件返回1，目录返回-1
        这样任意两个文件类型编号相乘，负数就代表不匹配
        """
        if self.is_file():
            return 1
        elif self.is_dir():
            return -1
        else:
            return 0

    def mtime(self):
        """ 文件的修改时间 """
        # windows会带小数，linux使用%Ts只有整数部分。
        # 这里不用四舍五入，取整数部分就是对应的。
        return int(os.stat(self).st_mtime)

    def size(self, *, human_readable=False):
        """ 获取文件/目录的大小 """
        if self.is_file():
            sz = os.path.getsize(self)
        elif self.is_dir():
            sz = sum([os.path.getsize(p) for p in self.rglob('*') if p.is_file()])
        else:
            sz = 0

        if human_readable:
            return humanfriendly.format_size(sz, binary=True)
        else:
            return sz

    def sub_rel_paths(self, mtime=False):
        """ 返回self目录下所有含递归的文件、目录，存储相对路径，as_posix
        当带有mtime参数时，会返回字典，并附带返回mtime的值

        主要用于 scp 同步目录数据时，对比目录下文件情况
        """

        if mtime:
            res = {}
            for p in self.glob('**/*'):
                res[p.relative_to(self).as_posix()] = p.mtime()
        else:
            res = set()
            for p in self.glob('**/*'):
                res.add(p.relative_to(self).as_posix())
        return res

    def relpath(self, ref_dir) -> 'XlPath':
        r""" 当前路径，相对于ref_dir的路径位置

        >>> File('C:/a/b/c.txt').relpath('C:/a/')
        'b/c.txt'
        >>> File('C:/a/b\\c.txt').relpath('C:\\a/')
        'b/c.txt'

        >> File('C:/a/b/c.txt').relpath('D:/')  # ValueError
        """
        return XlPath(os.path.relpath(self, str(ref_dir)))

    def as_windows_path(self):
        """ 返回windows风格的路径，即使用\\分隔符 """
        return self.as_posix().replace('/', '\\')

    def __contains__(self, item):
        """ 判断item的路径是否是在self里的，不考虑item是目录、文件，还是不存在文件的路径
        以前缀判断为准，有需要的话请自行展开resolve、expanduser再判断
        """
        if not self.is_file():
            # 根据操作系统判断路径分隔符
            separator = '\\' if os.name == 'nt' else '/'

            # 获取路径对象的字符串形式
            abs_path_str = str(self) + separator
            item_str = str(XlPath(item))

            # 判断路径字符串是否包含相对路径字符串
            return item_str.startswith(abs_path_str) or abs_path_str == item_str

    def get_total_lines(self, encoding='utf-8', skip_blank=False):
        """ 统计文件的行数（注意会统计空行，所以在某些场合可能与预期理解的条目数不太一致）

        :param str encoding: 文件编码，默认为'utf-8'
        :param bool skip_blank: 是否跳过空白行，默认为True
        :return: 文件的行数
        """
        line_count = 0
        with open(self, 'r', encoding=encoding) as file:
            for line in file:
                if skip_blank and not line.strip():  # 跳过空白行
                    continue
                line_count += 1
        return line_count

    def yield_line(self, start=0, end=None, step=1, batch_size=None, encoding='utf-8'):
        """ 返回指定区间的文件行

        :param int start: 起始行，默认为0
        :param int end: 结束行，默认为None（读取到文件末尾）
        :param int step: 步长，默认为1
        :param int batch_size: 每批返回的行数，如果为None，则逐行返回
        """
        total_lines = None  # 使用局部变量缓存总行数
        # 处理负索引
        if start < 0 or (end is not None and end < 0):
            total_lines = total_lines or self.get_total_lines()
            if start < 0:
                start = total_lines + start
            if end is not None and end < 0:
                end = total_lines + end

        with open(self, 'r', encoding=encoding) as file:
            iterator = islice(file, start, end, step)
            while True:
                batch = list(islice(iterator, batch_size))
                if not batch:
                    break
                batch = [line.rstrip('\n') for line in batch]  # 删除每行末尾的换行符
                if batch_size is None:
                    yield from batch
                else:
                    yield batch

    def split_to_dir(self, lines_per_file, dst_dir=None, encoding='utf-8',
                     filename_template="_{index}{suffix}"):
        """ 将文件按行拆分到多个子文件中

        :param int lines_per_file: 打算拆分的每个新文件的行数
        :param str dst_dir: 目标目录，未输入的时候，输出到同stem名的目录下
        :param str filename_template: 文件名模板，可以包含 {stem}, {index} 和 {suffix} 占位符
        :return list: 拆分的文件路径列表
            拆分后文件名类似如下： 01.jsonl, 02.jsonl, ...
        """
        # 1 检查输入参数
        if dst_dir is None:
            # 如果未提供目标目录，则拆分的文件保存到当前工作目录
            dst_dir = self.parent / f"{self.stem}"
        else:
            # 如果提供了目标目录，将拆分的文件保存到目标目录
            dst_dir = XlPath(dst_dir)

        if dst_dir.is_dir():
            raise FileExistsError(f"目标目录已存在，若确定要重置目录，请先删除目录：{dst_dir}")

        dst_dir.mkdir(parents=True, exist_ok=True)

        # 2 拆分文件
        split_files = []  # 用于保存拆分的文件路径
        outfile = None
        filename_format = "{:04d}"
        outfile_index = 0
        line_counter = 0
        suffix = self.suffix

        with open(self, 'r', encoding=encoding) as f:
            for line in f:
                if line_counter % lines_per_file == 0:
                    if outfile is not None:
                        outfile.close()
                    outfile_path = dst_dir / f"{self.stem}_{filename_format.format(outfile_index)}{suffix}"
                    outfile = open(outfile_path, 'w', encoding='utf-8')
                    split_files.append(outfile_path)  # 先占位，后面再填充
                    outfile_index += 1
                outfile.write(line)
                line_counter += 1

        if outfile is not None:
            outfile.close()

        # 3 重新设置文件名的对齐宽度
        new_filename_format = "{:0" + str(len(str(len(split_files)))) + "d}"
        for i, old_file in enumerate(split_files):
            new_name = dst_dir / filename_template.format(stem=self.stem,
                                                          index=new_filename_format.format(i),
                                                          suffix=suffix)
            os.rename(old_file, new_name)
            split_files[i] = new_name

        # 返回拆分的文件路径列表
        return split_files

    def merge_from_files(self, files,
                         ignore_empty_lines_between_files=False,
                         encoding='utf-8'):
        """ 将多个文件合并到一个文件中

        :param list files: 要合并的文件列表
        :param bool ignore_empty_lines_between_files: 是否忽略文件间的空行
        :param str encoding: 文件编码，默认为'utf-8'
        :return XlPath: 合并后的文件路径
        """
        # 合并文件
        prev_line_end_with_newline = True  # 记录上一次text的最后一个字符是否为'\n'
        with open(self, 'w', encoding=encoding) as outfile:
            for i, file in enumerate(files):
                file = XlPath(file)
                text = file.read_text(encoding=encoding)
                if ignore_empty_lines_between_files:
                    text = text.rstrip('\n')
                if i > 0 and not prev_line_end_with_newline and text != '':
                    outfile.write('\n')
                outfile.write(text)
                prev_line_end_with_newline = text.endswith('\n')

    def merge_from_dir(self, src_dir, filename_template="_{index}{suffix}", encoding='utf-8'):
        """ 将目录中的多个文件合并到一个文件中

        :param str src_dir: 要合并的文件所在的目录
        :param str filename_template: 文件名模板，可以包含 {stem}, {index} 和 {suffix} 占位符
        :param str encoding: 文件编码，默认为'utf-8'
        :return XlPath: 合并后的文件路径
        """
        src_dir = XlPath(src_dir)
        stem = src_dir.name

        pattern = filename_template.format(stem=stem, index=r"(\d+)", suffix=".*")
        files = [file for file in src_dir.iterdir() if re.match(pattern, file.name)]  # 获取目录中符合模式的文件

        self.merge_from_files(files, ignore_empty_lines_between_files=True, encoding=encoding)

    def __1_read_write(self):
        """ 参考标准库的
        read_bytes、read_text
        write_bytes、write_text
        """
        pass

    def read_text(self, encoding='utf8', errors='strict', return_mode: bool = False):
        """
        :param encoding: 效率拷贝，默认是设成utf8，但也可以设成None变成自动识别编码
        """
        if not encoding:
            result = charset_normalizer.from_path(self, cp_isolation=('utf_8', 'gbk', 'utf_16'))
            best_match = result.best()
            s = str(best_match)
            encoding = best_match.encoding
        else:
            with open(self, 'r', encoding=encoding, errors=errors) as f:
                s = f.read()

        # 如果用\r\n作为换行符会有一些意外不好处理
        if '\r' in s:
            s = s.replace('\r\n', '\n')

        if return_mode:
            return s, encoding
        else:
            return s

    def read_text2(self):
        """ 智能识别编码的文本读取，这里收集了我见过的一些常见类型 """
        for encoding in ['utf8',
                         'gbk',
                         'gb18030',
                         'utf_16',
                         'cp932',  # 日文，Shift-JIS
                         'Big5',  # 繁体字，Big5
                         'big5hkscs',  # 繁体字
                         ]:
            try:
                content = self.read_text(encoding=encoding)
                return content, encoding
            except (UnicodeDecodeError, UnicodeError):
                continue

    def readlines_batch(self, batch_size, *, encoding='utf8'):
        """ 将文本行打包，每次返回一个批次多行数据

        python的io.IOBase.readlines有个hint参数，不是预期的读取几行的功能，所以这里重点是扩展了一个readlines的功能

        :param batch_size: 默认每次获取一行内容，可以设参数，每次返回多行内容
            如果遍历每次只获取一行，一般不用这个接口，直接对open得到的文件句柄f操作就行了

        :return: list[str]
            注意返回的每行str，末尾都带'\n'
            但最后一行视情况可能有\n，可能没有\n

        注，开发指南：不然扩展支持batch_size=-1获取所有数据
        """
        f = open(self, 'r', encoding=encoding)
        return chunked(f, batch_size)

    def write_text(self, data, encoding='utf8', mode='w', errors=None, newline=None):
        with open(self, mode, encoding=encoding, errors=errors, newline=newline) as f:
            return f.write(data)

    def write_text_unix(self, data, encoding='utf8', mode='w', errors=None, newline='\n'):
        with open(self, mode, encoding=encoding, errors=errors, newline=newline) as f:
            return f.write(data)

    def read_pkl(self):
        with open(self, 'rb') as f:
            return pickle.load(f)

    def write_pkl(self, data):
        with open(self, 'wb') as f:
            pickle.dump(data, f)

    def read_json(self, encoding='utf8', *, errors='strict', return_mode: bool = False):
        """

        Args:
            encoding: 可以主动指定编码，否则默认会自动识别编码
            return_mode: 默认只返回读取的数据
                开启后，得到更丰富的返回信息: data, encoding
                    该功能常用在需要自动识别编码，重写回文件时使用相同的编码格式
        Returns:

        """
        s, encoding = self.read_text(encoding=encoding, errors=errors, return_mode=True)
        try:
            data = ujson.loads(s)
        except ValueError:  # ujson会有些不太标准的情况处理不了
            data = json.loads(s)

        if return_mode:
            return data, encoding
        else:
            return data

    def write_json(self, data, encoding='utf8', **kwargs):
        with open(self, 'w', encoding=encoding) as f:
            DictTool.ior(kwargs, {'ensure_ascii': False})
            json.dump(data, f, **kwargs)

    def read_jsonl(self, encoding='utf8', max_items=None, *,
                   errors='strict', return_mode=0, batch_size=None):
        """ 从文件中读取JSONL格式的数据

        :param str encoding: 文件编码格式，默认为utf8
        :param str errors: 读取文件时的错误处理方式，默认为strict
        :param bool return_mode: 是否返回文件编码格式，默认为False
            0, 读取全量数据返回
            1，返回文件编码格式
        :param int max_items: 限制读取的条目数，默认为None，表示读取所有条目
        :param int batch_size:
            默认为None，表示一次性读取所有数据
            如果设置了数值，则会流式读取，常用于太大，超过内存大小等的jsonl文件读取
                注意如果设置了大小，只是底层每次一批读取的大小，但返回的data仍然是一维的数据格式迭代器
        :return: 返回读取到的数据列表，如果return_mode为True，则同时返回文件编码格式

        >> read_jsonl('data.jsonl', max_items=10)  # 读取前10条数据
        """
        if batch_size is None:
            s, encoding = self.read_text(encoding=encoding, errors=errors, return_mode=True)

            data = []
            # todo 这一步可能不够严谨，不同的操作系统文件格式不同。但使用splitlines也不太好，在数据含有NEL等特殊字符时会多换行。
            for line in s.split('\n'):
                if line:
                    try:  # 注意，这里可能会有数据读取失败
                        data.append(json.loads(line))
                    except json.decoder.JSONDecodeError:
                        pass
                # 如果达到了限制的条目数，就停止读取
                if max_items is not None and len(data) >= max_items:
                    break
        else:
            def get_data():
                for batch in self.yield_line(batch_size=batch_size, encoding=encoding):
                    for line in batch:
                        try:  # 注意，这里可能会有数据读取失败
                            yield json.loads(line)
                        except json.decoder.JSONDecodeError:
                            pass

            data = get_data()

        if return_mode:
            return data, encoding
        else:
            return data

    def write_jsonl(self, list_data, ensure_ascii=False, default=None, mode='w', errors=None):
        """ 由于这种格式主要是跟商汤这边对接，就尽量跟它们的格式进行兼容 """
        content = '\n'.join([json.dumps(x, ensure_ascii=ensure_ascii, default=default) for x in list_data])
        self.write_text_unix(content + '\n', mode=mode, errors=errors)

    def add_json_line(self, data, ensure_ascii=False, default=None, mode='a'):
        """ 在文件末尾添加一行JSON数据 """
        content = json.dumps(data, ensure_ascii=ensure_ascii, default=default)
        self.write_text_unix(content + '\n', mode=mode)

    def read_csv(self, encoding='utf8', *, errors='strict', return_mode: bool = False,
                 delimiter=',', quotechar='"', **kwargs):
        """
        :return:
            data，n行m列的list
        """
        import csv

        s, encoding = self.read_text(encoding=encoding, errors=errors, return_mode=True)
        data = list(csv.reader(s.splitlines(), delimiter=delimiter, quotechar=quotechar, **kwargs))

        if return_mode:
            return data, encoding
        else:
            return data

    def read_yaml(self, encoding='utf8', *, errors='strict', rich_return=False):
        s, encoding = self.read_text(encoding=encoding, errors=errors, return_mode=True)
        data = yaml.safe_load(s)

        if rich_return:
            return data, encoding
        else:
            return data

    def write_yaml(self, data, encoding='utf8', *, sort_keys=False, **kwargs):
        with open(self, 'w', encoding=encoding) as f:
            yaml.dump(data, f, sort_keys=sort_keys, **kwargs)

    def read_auto(self, *args, **kwargs):
        """ 根据文件后缀自动识别读取函数 """
        if self.is_file():  # 如果存在这样的文件，那就读取文件内容
            # 获得文件扩展名，并统一转成小写
            mode = self.suffix.lower()[1:]
            read_func = getattr(self, 'read_' + mode, None)
            if read_func:
                return read_func(*args, **kwargs)
            elif mode in ('jpg', 'png'):  # 常见的一些二进制数据
                return self.read_bytes()
            else:
                return self.read_text(*args, **kwargs)
        else:  # 非文件对象
            raise FileNotFoundError(f'{self} 文件不存在，无法读取。')

    def write_auto(self, data, *args, if_exists=None, **kwargs):
        """ 根据文件后缀自动识别写入函数 """
        self.parent.mkdir(exist_ok=True, parents=True)
        mode = self.suffix.lower()[1:]
        write_func = getattr(self, 'write_' + mode, None)
        if self.exist_preprcs(if_exists):
            if write_func:
                return write_func(data, *args, **kwargs)
            else:
                return self.write_text(str(data), *args, **kwargs)
        return self

    def __2_glob(self):
        """ 类型判断、glob系列 """
        pass

    def ____1_常用glob(self):
        pass

    def glob(self, pattern):
        # TODO 加正则匹配模式
        if isinstance(pattern, str):
            return super(XlPath, self).glob(pattern)
        elif isinstance(pattern, (list, tuple)):
            from itertools import chain
            ls = [super(XlPath, self).glob(x) for x in pattern]
            # 去重
            exists = set()
            files = []
            for f in chain(*ls):
                k = f.as_posix()
                if k not in exists:
                    exists.add(k)
                    files.append(f)
            return files

    def glob_files(self, pattern='*'):
        for f in self.glob(pattern):
            if f.is_file():
                yield f

    def rglob_files(self, pattern='*'):
        for f in self.rglob(pattern):
            if f.is_file():
                yield f

    def glob_dirs(self, pattern='*'):
        for f in self.glob(pattern):
            if f.is_dir():
                yield f

    def rglob_dirs(self, pattern='*'):
        for f in self.rglob(pattern):
            if f.is_dir():
                yield f

    def glob_stems(self):
        """ 按照文件的stem分组读取

        :return: 返回格式类似 {'stem1': [suffix1, suffix2, ...], 'stem2': [suffix1, suffix2, ...], ...}
        """
        from collections import defaultdict
        d = defaultdict(set)
        for f in self.glob_files():
            d[f.stem].add(f.suffix)
        return d

    def glob_suffixs(self):
        """ 判断目录下有哪些扩展名的文件 """
        suffixs = set()
        for f in self.glob_files():
            suffixs.add(f.suffix)
        return suffixs

    def ____2_定制glob(self):
        pass

    def is_image(self):
        return self.is_file() and filetype.is_image(self)

    def _glob_images(self, glob_func, pattern):
        """ 在满足glob规则基础上，后缀还必须是合法的图片格式后缀
        """
        suffixs = {'png', 'jpg', 'jpeg', 'bmp', 'webp', 'gif'}
        for f in glob_func(pattern):
            if f.is_file() and f.suffix[1:].lower() in suffixs:
                yield f

    def glob_images(self, pattern='*'):
        """ 按照文件后缀，获取所有图片文件 """
        return self._glob_images(self.glob, pattern)

    def rglob_images(self, pattern='*'):
        return self._glob_images(self.rglob, pattern)

    def ____3_xglob(self):
        """ xglob系列，都是按照文件的实际内容，进行类型检索遍历的 """

    def xglob_images(self, pattern='*'):
        """ 按照文件实际内容，获取所有图片文件 """
        for f in self.glob(pattern):
            if f.is_file() and filetype.is_image(f):
                yield f

    def xglob_videos(self, pattern='*'):
        """ 按照文件实际内容，获取所有图片文件 """
        for f in self.glob(pattern):
            if f.is_file() and filetype.is_video(f):
                yield f

    def xglob_archives(self, pattern='*'):
        """ 找出所有压缩文件

        只找通用意义上的压缩包，不找docx等这种形式的文件
        """
        for f in self.glob(pattern):
            if f.is_file() and filetype.is_archive(f):
                if f.suffix in {'.pdf', '.docx', '.xlsx', '.pptx'}:
                    continue
                yield f

    def __3_文件基础操作(self):
        """ 复制、删除、移动等操作 """
        pass

    def exist_preprcs(self, if_exists=None):
        """ 从旧版File机制复制过来的函数

        这个实际上是在做copy等操作前，如果目标文件已存在，需要预先删除等的预处理
        并返回判断，是否需要执行下一步操作

        有时候情况比较复杂，process无法满足需求时，可以用exist_preprcs这个底层函数协助

        :param if_exists:
            None: 不做任何处理，直接运行，依赖于功能本身是否有覆盖写入机制
            'error': 如果要替换的目标文件已经存在，则报错
            'replace': 把存在的文件先删除
                本来是叫'delete'更准确的，但是考虑用户理解，
                    一般都是用在文件替换场合，叫成'delete'会非常怪异，带来不必要的困扰、误解
                所以还是决定叫'replace'
            'skip': 不执行后续功能
            'backup': 先做备份  （对原文件先做一个备份）
        """
        need_run = True
        if self.exists():
            if if_exists is None:
                return need_run
            elif if_exists == 'error':
                raise FileExistsError(f'目标文件已存在： {self}')
            elif if_exists == 'replace':
                self.delete()
            elif if_exists == 'skip':
                need_run = False
            elif if_exists == 'backup':
                self.backup(move=True)
            else:
                raise ValueError(f'{if_exists}')
        return need_run

    def copy(self, dst, if_exists=None):
        """ 用于一般的文件、目录拷贝

        不返回结果文件，是因为 if_exists 的逻辑比较特殊，比如skip的时候，这里不好实现返回的是最后的目标文件
        """
        if not self.exists():
            return

        dst = XlPath(dst)
        if dst.exist_preprcs(if_exists):
            if self.is_file():
                return shutil.copy2(self, dst)
            else:
                return shutil.copytree(self, dst)

    def move(self, dst, *, cross_disk=False, if_exists=None):
        """
        :param cross_disk: 是否可能涉及跨磁盘操作
        """
        if not self.exists():
            return self

        if cross_disk:  # 显式设置跨磁盘操作
            dst = self.copy(dst, if_exists=if_exists)
            self.delete()
            return dst

        try:
            dst = XlPath(dst)
            if self == dst:
                # 同一个文件，可能是调整了大小写名称
                if self.as_posix() != dst.as_posix():
                    tmp = self.tempfile(dir=self.parent)  # self不一定是file，也可能是dir，但这个名称通用
                    self.rename(tmp)
                    self.delete()
                    tmp.rename(dst)
            elif dst.exist_preprcs(if_exists):
                self.rename(dst)
        except OSError:
            # 有可能是跨磁盘操作，这个时候就只能先拷贝再删除了
            dst = self.copy(dst, if_exists=if_exists)
            self.delete()
        return dst

    def rename2(self, new_name, if_exists=None):
        """ 相比原版的rename，搞了更多骚操作，但性能也会略微下降，所以重写一个功能名

        240416周二12:49，这个接口将真的只做重命名，不做移动！所以将会不再支持dst中出现"/"路径配置
        """
        if not self.exists():
            return self

        if not isinstance(new_name, str):
            raise ValueError(f'rename2只能做重命名操作，目标路径必须是一个str')
        elif '/' in new_name:
            raise ValueError(f'rename2只能做重命名操作，目标路径中不能包含"/"')
        elif '\\' in new_name:
            raise ValueError(f'rename2只能做重命名操作，目标路径中不能包含"\\"')

        if self.name == new_name:  # 没有修改名称，跟原来相同
            return self

        dst = self.parent / new_name
        if self == dst:
            # 同一个文件，可能是调整了大小写名称
            if self.as_posix() != dst.as_posix():
                tmp = self.tempfile(dir=self.parent)  # self不一定是file，也可能是dir，但这个名称通用
                self.rename(tmp)
                self.delete()
                tmp.rename(dst)
        elif dst.exist_preprcs(if_exists):
            self.rename(dst)
        return dst

    def rename_stem(self, stem, if_exists=None):
        """ 重命名文件的stem """
        return self.rename2(stem + self.suffix, if_exists)

    def rename_suffix(self, suffix, if_exists=None):
        """ 重命名文件的suffix """
        return self.rename2(self.stem + suffix, if_exists)

    def delete(self):
        if self.is_file():
            os.remove(self)
        elif self.is_dir():
            shutil.rmtree(self)

    def backup(self, tail=None, if_exists='replace', move=False):
        r""" 对文件末尾添加时间戳备份，也可以使用自定义标记tail

        :param tail: 自定义添加后缀
            tail为None时，默认添加特定格式的时间戳
        :param if_exists: 备份的目标文件名存在时的处理方案
            这个概率非常小，真遇到，先把已存在的删掉，重新写入一个是可以接受的
        :param move: 是否删除原始文件

        # TODO：有个小bug，如果在不同时间实际都是相同一个文件，也会被不断反复备份
        #    如果想解决这个，就要读取目录下最近的备份文件对比内容了
        """
        from datetime import datetime

        # 1 判断自身文件是否存在
        if not self:
            return None

        # 2 计算出新名称
        if not tail:
            tail = datetime.fromtimestamp(self.mtime()).strftime(' %y%m%d-%H%M%S')  # 时间戳
        name, ext = os.path.splitext(str(self))
        dst = name + tail + ext

        # 3 备份就是特殊的copy操作
        if move:
            return self.move(dst, if_exists)
        else:
            return self.copy(dst, if_exists)

    def __4_重复文件相关功能(self):
        """ 检查目录里的各种文件情况 """

    def glob_repeat_files(self, pattern='*', *, sort_mode='count', print_mode=False,
                          files=None, hash_func=None):
        """ 返回重复的文件组

        :param files: 直接指定候选文件清单，此时pattern默认失效
        :param hash_func: hash规则，默认使用etag规则
        :param sort_mode:
            count: 按照重复的文件数量从多到少排序
            size: 按照空间总占用量从大到小排序
        :return: [(etag, files, per_file_size), ...]
        """
        # 0 文件清单和hash方法
        if files is None:
            files = list(self.glob_files(pattern))

        if hash_func is None:
            def hash_func(f):
                return get_etag(str(f))

        # 1 获取所有etag，这一步比较费时
        hash2files = defaultdict(list)

        for f in tqdm(files, desc='get etags', disable=not print_mode):
            etag = hash_func(f)
            hash2files[etag].append(f)

        # 2 转格式，排序
        hash2files = [(k, vs, vs[0].size()) for k, vs in hash2files.items() if len(vs) > 1]
        if sort_mode == 'count':
            hash2files.sort(key=lambda x: (-len(x[1]), -len(x[1]) * x[2]))
        elif sort_mode == 'size':
            hash2files.sort(key=lambda x: (-len(x[1]) * x[2], -len(x[1])))

        # 3 返回每一组数据
        return hash2files

    def delete_repeat_files(self, pattern='*', *, sort_mode='count', print_mode=True, debug=False,
                            files=None, hash_func=None):
        """
        :param debug:
            True，只是输出检查清单，不做操作
            False, 保留第1个文件，删除其他文件
            TODO，添加其他删除模式
        :param print_mode:
            0，不输出
            'str'，普通文本输出（即返回的msg）
            'html'，TODO 支持html富文本显示，带超链接
        """
        from humanfriendly import format_size

        def printf(*args, **kwargs):
            if print_mode == 'html':
                raise NotImplementedError
            elif print_mode:
                print(*args, **kwargs)
            msg.append(' '.join(args))

        fmtsize = lambda x: format_size(x, binary=True)

        msg = []
        files = self.glob_repeat_files(pattern, sort_mode=sort_mode, print_mode=print_mode,
                                       files=files, hash_func=hash_func)
        for i, (etag, files, _size) in enumerate(files, start=1):
            n = len(files)
            printf(f'{i}、{etag}\t{fmtsize(_size)} × {n} ≈ {fmtsize(_size * n)}')

            for j, f in enumerate(files, start=1):
                if debug:
                    printf(f'\t{f.relpath(self)}')
                else:
                    if j == 1:
                        printf(f'\t{f.relpath(self)}')
                    else:
                        f.delete()
                        printf(f'\t{f.relpath(self)}\tdelete')
            if print_mode:
                printf()

        return msg

    def check_repeat_files(self, pattern='**/*', **kwargs):
        if 'debug' not in kwargs:
            kwargs['debug'] = True
        return self.delete_repeat_files(pattern, **kwargs)

    def check_repeat_name_files(self, pattern='**/*', **kwargs):
        if 'hash_func' not in kwargs:
            kwargs['hash_func'] = lambda p: p.name.lower()
        return self.check_repeat_files(pattern, **kwargs)

    def __5_文件后缀相关功能(self):
        """ 检查目录里的各种文件情况 """

    def refine_files_suffix(self, pattern='*', *, print_mode=True, if_exists=None, debug=False):
        """ 优化文件的后缀名 """
        j = 1
        for i, f1 in enumerate(self.glob_files(pattern), start=1):
            suffix1 = f1.suffix

            suffix2 = suffix1.lower()
            if suffix2 == '.jpeg':
                suffix2 = '.jpg'

            if suffix1 != suffix2:
                f2 = f1.with_suffix(suffix2)

                if print_mode:
                    print(f'{i}/{j} {f1} -> {suffix2}')

                if not debug:
                    f1.rename2(f2, if_exists=if_exists)
                j += 1

    def _check_faker_suffix(self, file_list):
        """ 检查文件扩展名是否匹配实际内容类型，并迭代文件列表进行处理。

        :param file_list: 文件列表
        :return: 迭代器，产生文件路径和相应的信息
        """
        for file_path in file_list:
            t = filetype.guess(file_path)
            if not t:
                continue

            ext = '.' + t.extension
            ext0 = file_path.suffix
            if ext == ext0:
                continue
            elif ext == '.xls' and ext0 == '.et':
                continue

            if ext0 in ('.docx', '.xlsx', '.pptx', '.xlsm'):
                ext0 = '.zip'
            elif ext0 in ('.JPG', '.jpeg'):
                ext0 = '.jpg'

            if ext != ext0:
                yield file_path, ext

    def xglob_faker_suffix_files(self, pattern='*'):
        """ 检查文件扩展名是不是跟实际内容类型不匹配，有问题

        注：推荐先运行 refine_files_suffix，本函数是大小写敏感，并且不会区分jpeg和jpg
        """
        return self._check_faker_suffix(self.glob_files(pattern))

    def xglob_faker_suffix_images(self, pattern='*'):
        """ 只检查原本就是图片名称标记的文件的相关数据正误
        """
        return self._check_faker_suffix(self.glob_images(pattern))

    def rename_faker_suffix_files(self, pattern='*', *, print_mode=True, if_exists=None, debug=False):
        for i, (f1, suffix2) in enumerate(self.xglob_faker_suffix_files(pattern), start=1):
            if print_mode:
                print(f'{i}、{f1} -> {suffix2}')

            if not debug:
                f2 = f1.with_suffix(suffix2)
                f1.rename2(f2, if_exists=if_exists)

    def __6_文件夹分析诊断(self):
        pass

    def check_size(self, return_mode='str'):
        import pandas as pd

        msg = []
        file_sizes = {}  # 缓存文件大小，避免重复计算
        suffix_counts = Counter()
        suffix_sizes = defaultdict(int)

        dir_count, file_count = 0, 0
        for root, dirs, files in os.walk(self):
            dir_count += len(dirs)
            file_count += len(files)
            for file in files:
                file_size = os.path.getsize(os.path.join(root, file))
                file_sizes[(root, file)] = file_size

                _, suffix = os.path.splitext(file)
                suffix_counts[suffix] += 1
                suffix_sizes[suffix] += file_size

        sz = human_readable_size(sum(file_sizes.values()))
        # 这里的目录指"子目录"数，不包含self
        msg.append(f'一、目录数：{dir_count}，文件数：{file_count}，总大小：{sz}')

        data = []
        for suffix, count in suffix_counts.most_common():
            size = suffix_sizes[suffix]
            data.append([suffix, count, size])
        data.sort(key=lambda x: (-x[2], -x[1], x[0]))  # 先按文件数，再按文件名排序
        df = pd.DataFrame(data, columns=['suffix', 'count', 'size'])
        df['size'] = [human_readable_size(x) for x in df['size']]
        df.reset_index(inplace=True)
        df['index'] += 1
        msg.append('\n二、各后缀文件数')
        msg.append(df.to_string(index=False))

        if return_mode == 'str':
            return '\n'.join(msg)
        elif return_mode == 'list':
            return msg
        else:
            return msg

    def check_summary(self, print_mode=True, return_mode=False, **kwargs):
        if self.is_dir():
            res = self._check_dir_summary(print_mode, **kwargs)
        elif self.is_file():
            res = self._check_file_summary(print_mode, **kwargs)
        else:
            res = '文件不存在'
            print(res)

        if return_mode:
            return res

    def _check_file_summary(self, print_mode=True, **kwargs):
        """ 对文件进行通用的状态检查

        :param bool print_mode: 是否将统计信息打印到控制台
        :return dict: 文件的统计信息
        """
        file_summary = {}

        # 文件大小
        file_summary['文件大小'] = self.size(human_readable=True)

        # 文件行数
        file_summary['文件行数'] = self.get_total_lines()

        # 文件修改时间
        mod_time_str = datetime.datetime.fromtimestamp(self.mtime()).strftime('%Y-%m-%d %H:%M:%S')
        file_summary['修改时间'] = mod_time_str

        # 如果print_mode为True，则将统计信息打印到控制台
        if print_mode:
            for key, value in file_summary.items():
                print(f"{key}: {value}")

        return file_summary

    def _check_dir_summary(self, print_mode=True, hash_func=None, run_mode=31):
        """ 对文件夹情况进行通用的状态检查

        :param hash_func: 可以传入自定义的hash函数，用于第四块的重复文件运算
            其实默认的get_etag就没啥问题，只是有时候为了性能考虑，可能会传入一个支持，提前有缓存知道etag的函数
        :param int run_mode: 只运行编号内的功能
        """
        if not self.is_dir():
            return ''

        def printf(s):
            if print_mode:
                print(s)
            msg.append(s)

        # 一 目录大小，二 各后缀文件大小
        msg = []
        if run_mode & 1:  # 1和2目前是绑定一起运行的
            printf('【' + self.as_posix() + '】目录检查')
            printf('\n'.join(self.check_size('list')))

        # 三 重名文件
        if run_mode & 2:
            printf('\n三、重名文件（忽略大小写，跨目录检查name重复情况）')
            printf('\n'.join(self.check_repeat_name_files(print_mode=False)))

        # 四 重复文件
        if run_mode & 4:
            printf('\n四、重复文件（etag相同）')
            printf('\n'.join(self.check_repeat_files(print_mode=False, hash_func=hash_func)))

        # 五 错误扩展名
        if run_mode & 8:
            printf('\n五、错误扩展名')
            for i, (f1, suffix2) in enumerate(self.xglob_faker_suffix_files('**/*'), start=1):
                printf(f'{i}、{f1.relpath(self)} -> {suffix2}')

        # 六 文件配对
        if run_mode & 16:
            printf(
                '\n六、文件配对（检查每个目录里stem名称是否配对，列出文件组成不单一的目录结构，请重点检查落单未配对的情况）')
            prompt = False
            for root, dirs, files in os.walk(self):
                suffix_counts = defaultdict(list)
                for file in files:
                    stem, suffix = os.path.splitext(file)
                    suffix_counts[stem].append(suffix)
                suffix_counts = {k: tuple(sorted(v)) for k, v in suffix_counts.items()}
                suffix_counts2 = {v: k for k, v in suffix_counts.items()}  # 反向存储，如果有重复v会进行覆盖
                ct = Counter(suffix_counts.values())
                if len(ct.keys()) > 1:
                    printf(root)
                    for k, v in ct.most_common():
                        tag = f'\t{k}: {v}'
                        if v == 1:
                            tag += f'，{suffix_counts2[k]}'
                        if len(k) > 1 and not prompt:
                            tag += f'\t标记注解：有{v}组stem相同文件，配套有{k}这些后缀。其他标记同理。'
                            prompt = True
                        printf(tag)

        return '\n'.join(msg)

    def __7_目录复合操作(self):
        """ 比较高级的一些目录操作功能 """

    def delete_empty_subdir(self, recursive=True, topdown=False):
        """ 删除指定目录下的所有空目录。

        :param recursive: 是否递归删除所有子目录。
        :param topdown: 是否从顶部向下遍历目录结构。
            默认False，要先删除内部目录，再删除外部目录。
        """

        def _delete_empty_dirs(dir_path):
            for dir_name in os.listdir(dir_path):
                dir_fullpath = os.path.join(dir_path, dir_name)
                if os.path.isdir(dir_fullpath):
                    # 递归删除子目录中的空目录
                    _delete_empty_dirs(dir_fullpath)
                    if not os.listdir(dir_fullpath):
                        # 删除空目录
                        os.rmdir(dir_fullpath)

        if not recursive:
            for dirname in os.listdir(self):
                dir_fullpath = os.path.join(self, dirname)
                if os.path.isdir(dir_fullpath) and not os.listdir(dir_fullpath):
                    os.rmdir(dir_fullpath)
        else:
            for dirpath, dirnames, filenames in os.walk(self, topdown=topdown):
                for dirname in dirnames:
                    dir_fullpath = os.path.join(dirpath, dirname)
                    if not os.listdir(dir_fullpath):
                        os.rmdir(dir_fullpath)
                # 如果不递归删除子目录，则直接跳过
                if not recursive:
                    break

    def flatten_directory(self, *, clear_empty_subdir=True):
        """ 将子目录的文件全部取出来，放到外面的目录里

        :param clear_empty_subdir: 移除文件后，删除空子目录
        """
        # 1 检查是否有重名文件，如果有重名文件则终止操作
        if msg := self.check_repeat_name_files(print_mode=False):
            print('\n'.join(msg))
            raise ValueError('有重名文件，终止操作')

        # 2 操作
        self._flatten_directory_recursive(self, clear_empty_subdir)

    def _flatten_directory_recursive(self, current_dir, clear_empty_subdir):
        for name in os.listdir(current_dir):
            path = os.path.join(current_dir, name)

            if os.path.isdir(path):
                # If it's a directory, recursively flatten it
                self._flatten_directory_recursive(path, clear_empty_subdir)
                if clear_empty_subdir:
                    shutil.rmtree(path)
            elif os.path.isfile(path):
                # If it's a file, move it to the top-level directory
                destination_path = os.path.join(self, name)
                shutil.move(path, destination_path)

    def _nest_directory_core(self, file_names, min_files_per_batch=None, groupby=None):
        """ 核心方法：将文件列表按照指定规则进行分组

        :param file_names: 文件列表
        :param min_files_per_batch: 每个批次最少包含的文件数，默认为 None，即不限制最少文件数
        :param groupby: 分组函数，用于指定按照哪个属性进行分组，默认为 None
        :return: 分组结果列表
        """
        from pyxllib.algo.pupil import Groups, natural_sort

        if groupby is None:
            groupby = lambda p: p.stem.lower()
        file_groups = Groups.groupby(file_names, groupby).data

        if min_files_per_batch is None:
            min_files_per_batch = 1

        result_groups, current_group = [], []
        for stem in natural_sort(file_groups.keys()):
            current_group += file_groups[stem]
            if len(current_group) >= min_files_per_batch:
                result_groups.append(current_group)
                current_group = []
        if current_group:
            result_groups.append(current_group)

        return result_groups

    def nest_directory(self, min_files_per_batch=None, groupby=None, batch_name=None, bias=0, tail_limit=None):
        """ 将直接子文件按照一定规则拆分成多个batch子目录
        注意这个功能和flatten_directory是对称的，所以函数名也是对称的

        :param min_files_per_batch: 每个batch最少含有的文件数
            None，相当于int=1的效果
            int, 如果输入一个整数，则按照这个数量约束分成多个batch
        :param groupby: 默认会把stem.lower()相同的强制归到一组
            def groupby(p: XlPath) -> 分组用的key，相同key会归到同一组
            这个不仅用于分组，返回的字符串，也会作为字典序排序的依据，如果想用自然序，记得加natural_sort_key进行转换
        :param batch_name: 设置batch的名称，默认 'batch{}'
        :param bias: 希望用不到这个参数，只有中途出bug，需要继续处理的时候，用来自动增加编号
            注意默认就是从1开始编号的，比如bias设成8的话，实际是从9开始编号的
        :param tail_limit: 限制数量少于多少的batch，合并到上一个batch中
        """
        from pyxllib.algo.pupil import Groups

        # 1 按stem分组，确定分组数
        file_names = list(self.glob_files('*'))
        if groupby is None:
            groupby = lambda p: p.stem.lower()
        file_groups = Groups.groupby(file_names, groupby).data

        if min_files_per_batch is None:
            min_files_per_batch = 1

        # 2 将文件组合并为多个分组，每组至少包含 min_files_per_batch 个文件
        result_groups, current_group = [], []
        for stem in sorted(file_groups.keys()):  # 注意这里需要对取到的key按照自然序排序
            current_group += file_groups[stem]
            if len(current_group) >= min_files_per_batch:
                result_groups.append(current_group)
                current_group = []
        if current_group:
            if tail_limit is None:
                tail_limit = min_files_per_batch // 10

            if len(current_group) < tail_limit and result_groups:
                result_groups[-1] += current_group
            else:
                result_groups.append(current_group)

        # 3 整理实际的文件
        if batch_name is None:
            group_num = len(result_groups)
            width = len(str(group_num))
            batch_name = f'batch{{:0{width}}}'
        for i, group in enumerate(result_groups, start=1):
            d = self / batch_name.format(i + bias)
            d.mkdir(exist_ok=True)
            for f in group:
                f.move(d / f.name)

    def select_file(self, pos_filter=None, *, neg_filter=None):
        from pyxllib.file.specialist.dirlib import Dir

        d = Dir(self)
        if pos_filter is None:
            # 基于filesmatch的底层来实现，速度会比较慢一些，但功能丰富，不用重复造轮子
            d = d.select('**/*', type_='file')
        else:
            d = d.select(pos_filter, type_='file')
        if neg_filter is not None:
            d = d.exclude(neg_filter)

        files = [(self / f) for f in d.subs]
        return files

    def copy_file_filter(self, dst, pos_filter=None, *, neg_filter=None, if_exists=None):
        """ 只能用于目录，在复制文件的时候，进行一定的筛选，而不是完全拷贝

        :param dst: 目标目录
        :param pos_filter: 对文件的筛选规则
            其实常用的就'*.json'这种通配符
            支持自定义函数，def pos_filter(p: XlPath)
            参数详细用法参考 filesmatch，注意这里筛选器只对file启用，不会对dir启用，否则复制逻辑会非常乱
        :param neg_filter:

        注意！这个功能还是有点特殊的，不建议和XlPath.copy的接口做合并。

        正向、反向两个过滤器可以组合使用，逻辑上是先用正向获取全部文件，然后扣除掉反向。
        正向默认全选，反向默认不扣除。

        注意：这样过滤后，空目录不会被拷贝。
            如果有拷贝目录的需求，请自己另外写逻辑实现。
            如果只是拷贝空目录结构的需求，可以使用 copy_dir_structure

        >> p = XlPath('build')
        >> p.copy_filter('build2', '*.toc')  # 复制直接子目录下的所有toc文件
        >> p.copy_filter('build2', '**/*.toc')  # 复制所有toc文件
        >> p.copy_filter('build2', lambda p: p.suffix == '.toc')  # 复制所有toc文件
        """
        files = self.select_file(pos_filter, neg_filter=neg_filter)

        dst = XlPath(dst)
        for f in files:
            dst2 = dst / f
            dst2.parent.mkdir(exist_ok=True)
            (self / f).copy(dst2, if_exists=if_exists)

    def copy_dir_structure(self, dst):
        """ 只复制目录结构，不复制文件内容 """
        dst = XlPath(dst)
        for root, dirs, _ in os.walk(self):
            # 构造目标路径
            dst_dir = dst / XlPath(root).relative_to(self)
            # 创建目录
            dst_dir.mkdir(parents=True, exist_ok=True)

    # 无法选定文件
    def _move_selectable(self, dst_dir, *, print_mode=False):
        """ 目录功能，将目录下可选中的文件移动到目标目录

        1、要理解这个看似有点奇怪的功能，需要了解一个背景，在数据处理中，可能会拿到超长文件名的文件，
            这种在windows平台虽然手动可以操作，但在代码中，会glob不到，强制指定也会说文件不存在
        2、为了解决这类文件问题，一般需要对其进行某种规则的重命名。因为linux里似乎不会限制文件名长度，所以要把这些特殊文件打包到linux里处理。
        3、因为这些文件本来就无法被选中，所以只能反向操作，将目录下的可选中文件移动到目标目录。
        """
        for p in tqdm(self.glob('*'), disable=not print_mode):
            # 231211周一16:03 一般本来就glob不到，现在的p就是存在的，但是可能以防万一加的捕捉，我现在也不敢删
            if p.is_file():
                try:
                    p.move(dst_dir / p.name)
                except FileNotFoundError:
                    continue

    def move_unselectable(self, dst_dir, *, print_mode=False):
        """ 见_move_selectable，因为无法对这些特殊文件进行移动
        所以这里只是对_move_selectable的封装，中间通过文件重命名，来伪造移动了无法选中文件的操作效果
        """
        tempdir = self.create_tempdir_path(dir=self.parent)
        tempdir.mkdir(exist_ok=True)
        self._move_selectable(tempdir, print_mode=print_mode)
        self.rename2(dst_dir)
        tempdir.move(self)

    def rename_stem_until_not_exists(self):
        """ 比较高级的一个操作，会按照某种规则不断重命名，直到是当前并不存在的文件名，常用在文件拷贝避免重名冲突等场景

        todo 写个支持自定义规则的输出参数？
        todo 有个首次要不要判断exists的问题，可能跟不同的业务场景有关，要思考怎么设计更好...
        """

        def add_stem(m):
            a = int(m.group(1)) + 1
            return f'({a})'

        file = self
        while file.exists():
            stem = file.stem
            m = re.search(r'\(\d+\)$', stem)
            if m:  # 已经有目标范式的编号，继续累加
                stem = re.sub(r'\((\d+)\)$', add_stem, stem)
            else:  # 还没有编号的，直接从'2'开始编号
                stem += ' (2)'
            file = self.with_stem(stem.strip())  # 忽略最后空白，这个很容易出问题

        return file


class StreamJsonlWriter:
    """ 流式存储，主要用于存储文本化、jsonl格式数据 """

    def __init__(self, file_path, batch_size=2000, *,
                 delete_origin_file=False, json_default=str):
        self.file_path = XlPath(file_path)
        self.cache_text_lines = []
        self.batch_size = batch_size
        self.total_lines = 0

        self.delete_origin_file = delete_origin_file
        self.json_default = json_default

    def append_line(self, line):
        self.append_lines([line])

    def append_lines(self, data):
        """
        :param list data: 添加一组数据
        """
        for x in data:
            if isinstance(x, str):
                self.cache_text_lines.append(x)
            else:
                self.cache_text_lines.append(json.dumps(x, ensure_ascii=False,
                                                        default=self.json_default))
        if len(self.cache_text_lines) >= self.batch_size:
            self.flush()

    def flush(self):
        """ 刷新，将当前缓存写入文件 """
        if self.cache_text_lines:
            if self.total_lines == 0 and self.delete_origin_file:  # 第一次写入时，删除旧缓存文件
                self.file_path.delete()

            self.total_lines += len(self.cache_text_lines)
            self.file_path.parent.mkdir(exist_ok=True, parents=True)
            with open(self.file_path, 'a', encoding='utf8') as f:
                f.write('\n'.join(self.cache_text_lines) + '\n')
            self.cache_text_lines = []


def demo_file():
    """ File类的综合测试"""
    temp = tempfile.gettempdir()

    # 切换工作目录到临时文件夹
    os.chdir(temp)

    p = File('demo_path', temp)
    p.delete()  # 如果存在先删除
    p.ensure_parent()  # 然后再创建一个空目录

    print(File('demo_path', suffix='.py'))
    # F:\work\CreatorTemp\demo_path.py

    print(File('demo_path/', suffix='.py'))
    # F:\work\CreatorTemp\demo_path\tmp65m8mc0b.py

    # ...区别于None，会随机生成一个文件名
    print(File(..., temp))
    # F:\work\CreatorTemp\tmpwp4g1692

    # 可以在随机名称基础上，再指定文件扩展名
    print(File('', temp, suffix='.txt'))
    # F:\work\CreatorTemp\tmpimusjtu1.txt


def demo_file_rename():
    # 建一个空文件
    f1 = File('temp/a.txt', tempfile.gettempdir())
    # Path('F:/work/CreatorTemp/temp/a.txt')
    with open(str(f1), 'wb') as p: pass  # 写一个空文件

    f1.rename('A.tXt')  # 重命名
    # Path('F:/work/CreatorTemp/temp/a.txt')

    f1.rename('figs/b')  # 放到一个新的子目录里，并再次重命名
    # Path('F:/work/CreatorTemp/temp/figs/b')

    # TODO 把目录下的1 2 3 4 5重命名为5 4 3 2 1时要怎么搞？


class XlBytesIO(io.BytesIO):
    """ 自定义的字节流类，封装了struct_unpack操作

    https://www.yuque.com/xlpr/pyxllib/xlbytesio

    """

    def __init__(self, init_bytes):
        if isinstance(init_bytes, (File, str)):
            # with open的作用：可以用.read循序读入，而不是我的Path.read一口气读入。
            #   这在只需要进行局部数据分析，f本身又非常大的时候很有用。
            #   但是我这里操作不太方便等原因，还是先全部读入BytesIO流了
            init_bytes = File(init_bytes).read(mode='b')
        super().__init__(init_bytes)

    def unpack(self, fmt):
        return struct_unpack(self, fmt)

    def readtext(self, char_num, encoding='gbk', errors='ignore', code_length=2):
        """ 读取二进制流，将其解析为文本内容

        :param char_num: 字符数
        :param encoding: 所用编码，一般是gbk，因为如果是utf8是字符是变长的，不好操作
        :param errors: decode出现错误时的处理方式
        :param code_length: 每个字符占的长度
        :return: 文本内容
        """
        return self.read(code_length * char_num).decode(encoding, errors)


class PathGroups(Groups):
    """ 按stem文件名（不含后缀）分组的相关功能 """

    @classmethod
    def groupby(cls, files, key=lambda x: os.path.splitext(XlPath(x).as_posix())[0], ykey=lambda y: y.suffix[1:]):
        """
        :param files: 用Dir.select选中的文件、目录清单
        :param key: D:/home/datasets/textGroup/SROIE2019+/data/task3_testcrop/images/X00016469670
        :param ykey: ['jpg', 'json', 'txt']
        :return: dict
            1, task3_testcrop/images/X00016469670：['jpg', 'json', 'txt']
            2, task3_testcrop/images/X00016469671：['jpg', 'json', 'txt']
            3, task3_testcrop/images/X51005200931：['jpg', 'json', 'txt']
        """
        return super().groupby(files, key, ykey)

    def select_group(self, judge):
        """ 对于某一组，只要该组有满足judge的元素则保留该组 """
        data = {k: v for k, v in self.data.items() if judge(k, v)}
        return type(self)(data)

    def select_group_which_hassuffix(self, pattern, flags=re.IGNORECASE):
        def judge(k, values):
            for v in values:
                m = re.match(pattern, v, flags=flags)
                if m and len(m.group()) == len(v):
                    # 不仅要match满足，还需要整串匹配，比如jpg就必须是jpg，不能是jpga
                    return True
            return False

        return self.select_group(judge)

    def select_group_which_hasimage(self, pattern=r'jpe?g|png|bmp', flags=re.IGNORECASE):
        """ 只保留含有图片格式的分组数据 """
        return self.select_group_which_hassuffix(pattern, flags)

    def find_files(self, name, *, count=-1):
        """ 找指定后缀的文件

        :param name: 支持 '1.jpg', 'a/1.jpg' 等格式
        :param count: 返回匹配数量上限，-1表示返回所有匹配项
        :return: 找到第一个匹配项后返回
            找的有就返回XlPath对象
            找没有就返回None

        注意这个功能是大小写敏感的，如果出现大小写不匹配
            要么改文件名本身，要么改name的格式

        TODO 如果有大量的检索，这样每次遍历会很慢，可能要考虑构建后缀树来处理
        """
        ls = []
        stem, ext = os.path.splitext(name)
        stem = '/' + stem.replace('\\', '/')
        ext = ext[1:]
        for k, v in self.data.items():
            if k.endswith(stem) and ext in v:
                ls.append(XlPath.init(k, suffix=ext))
                if len(ls) >= count:
                    break
        return ls


def cache_file(file, make_data_func: Callable[[], Any] = None, *,
               mode='read_first',
               cache_time=None,
               **kwargs):
    """ 能将局部函数功能结果缓存进文件的功能

    输入的文件file如果存在则直接读取内容；
    否则用make_data_func生成，并且备份一个file文件

    :param file: 需要缓存的文件路径
    :param make_data_func: 如果文件不存在，则需要生成一份，要提供数据生成函数
        cache_file可以当装饰器用，此时不用显式指定该参数
    :param mode:
        read_first（默认）: 优先尝试从已有文件读取
        generate_first: 函数生成优先
    :param cache_time: 文件缓存时间，单位为秒，默认为None，表示始终使用缓存文件
        如果设置60，表示超过60秒后，需要重新优先从函数获得更新内容
    :param kwargs: 可以传递read、write支持的扩展参数
    :return: 读取到的数据
    """
    from datetime import datetime, timedelta
    from pyxllib.prog.pupil import format_exception

    def decorator(func):
        def wrapper(*args2, **kwargs2):

            f = XlPath.init(file, XlPath.tempdir())
            f.parent.mkdir(exist_ok=True, parents=True)

            # 1 优先看是不是需要先从文件读取数据
            if mode == 'read_first' and f.is_file():
                if cache_time is None:
                    return f.read_auto(**kwargs)

                current_time = datetime.now()
                last_modified = datetime.fromtimestamp(f.mtime())  # 获取文件的修改时间
                if not isinstance(cache_time, timedelta):
                    cache_time2 = timedelta(seconds=cache_time)
                else:
                    cache_time2 = cache_time

                if cache_time is None or (current_time - last_modified <= cache_time2):
                    return f.read_auto(**kwargs)

            # 2 如果需要重新生成数据，且没有已存在的保底文件
            if not f.is_file():
                data = func(*args2, **kwargs2)
                f.write_auto(data, **kwargs)
                return data

            # 3 需要重新生成，但是有保底文件
            try:
                data = func(*args2, **kwargs2)
                f.write_auto(data, **kwargs)
                return data
            except Exception as e:
                print(format_exception(e))
                return f.read_auto(**kwargs)

        return wrapper

    return decorator(make_data_func)() if make_data_func else decorator


class UsedRecords:
    """存储用户的使用记录到一个文件"""

    def __init__(self, filename, default_value=None, *, use_temp_root=False, limit_num=30):
        """ 记录存储文件

        :param filename: 文件路径与名称
        :param default_value:
        :param use_temp_root: 使用临时文件夹作为根目录
        :param limit_num: 限制条目上限
        """
        from os.path import join, dirname, basename, exists
        from pyxllib.text.specialist import ensure_content

        # 1 文件名处理
        if use_temp_root:
            dirname = join(os.getenv('TEMP'), 'code4101py_config')
            basename = basename(filename)
            fullname = join(dirname, basename)
        else:
            dirname = dirname(filename)
            basename = basename(filename)
            fullname = filename

        # 2 读取值
        if exists(fullname):
            ls = ensure_content(fullname).splitlines()
        else:
            ls = list(default_value)

        # 3 存储到类
        self.dirname = dirname
        self.basename = basename
        self.fullname = fullname
        self.ls = ls
        self.limit_num = limit_num

    def save(self):
        """保存记录文件"""
        File(self.dirname + '/').ensure_parent()
        File(self.fullname).write('\n'.join(self.ls), if_exists='replace')

    def add(self, s):
        """新增一个使用方法
        如果s在self.ls里，则把方法前置到第一条
        否则在第一条添加新方法

        如果总条数超过30要进行删减
        """
        if s in self.ls:
            del self.ls[self.ls.index(s)]

        self.ls = [s] + list(self.ls)

        if len(self.ls) > self.limit_num:
            self.ls = self.ls[:self.limit_num]

    def __str__(self):
        res = list()
        res.append(self.fullname)
        for t in self.ls:
            res.append(t)
        return '\n'.join(res)


def __5_filedfs():
    """
    对目录的遍历查看目录结构
    """


def file_generator(f):
    """普通文件迭代生成器
    :param f: 搜索目录
    """
    if os.path.isdir(f):
        try:
            dirpath, dirnames, filenames = myoswalk(f).__next__()
        except StopIteration:
            return []

        ls = filenames + dirnames
        ls = map(lambda x: dirpath + '/' + x, ls)
        return ls
    else:
        return []


def pyfile_generator(f):
    """py文件迭代生成器
    :param f: 搜索目录
    """
    if os.path.isdir(f):
        try:
            dirpath, dirnames, filenames = myoswalk(f).__next__()
        except StopIteration:
            return []
        filenames = list(filter(lambda x: x.endswith('.py'), filenames))
        ls = filenames + dirnames
        ls = map(lambda x: dirpath + '/' + x, ls)
        return ls
    else:
        return []


def texfile_generator(f):
    """tex  文件迭代生成器
    :param f: 搜索目录
    """
    if os.path.isdir(f):
        try:
            dirpath, dirnames, filenames = myoswalk(f).__next__()
        except StopIteration:
            return []

        filenames = list(filter(lambda x: x.endswith('.tex'), filenames))
        ls = filenames + dirnames
        ls = map(lambda x: dirpath + '/' + x, ls)
        return ls
    else:
        return []


def file_str(f):
    """
    :param f: 输入完整路径的文件夹或文件名
    :return: 返回简化的名称
        a/b     ==> <b>
        a/b.txt ==> b.txt
    """
    name = os.path.basename(f)
    if os.path.isdir(f):
        s = '<' + name + '>'
    else:
        s = name
    return s


def filedfs(root,
            child_generator=file_generator, select_depth=None, linenum=True,
            mystr=file_str, msghead=True, lsstr=None, show_node_type=False, prefix='\t'):
    """对文件结构的递归遍历
    注意这里的子节点生成器有对非常多特殊情况进行过滤，并不是通用的文件夹查看工具
    """
    from pyxllib.text.specialist import dfs_base

    if isinstance(child_generator, str):
        if child_generator == '.py':
            child_generator = pyfile_generator
        elif child_generator == '.tex':
            child_generator = texfile_generator
        else:
            raise ValueError

    return dfs_base(root, child_generator=child_generator, select_depth=select_depth, linenum=linenum,
                    mystr=mystr, msghead=msghead, lsstr=lsstr, show_node_type=show_node_type, prefix=prefix)


def genfilename(fd='.'):
    """生成一个fd目录下的文件名
    注意只是文件名，并未实际产生文件，输入目录是为了防止生成重名文件（以basename为标准的无重名）

    格式为：180827周一195802，如果出现重名，前面的6位记为数值d1，是年份+月份+日期的标签
        后面的6位记为数值d2，类似小时+分钟+秒的标签，但是在出现重名时，
        d2会一直自加1直到没有重名文件，所以秒上是可能会出现“99”之类的值的。
    """
    from datetime import datetime
    # 1 获取前段标签
    dt = datetime.now()
    weektag = '一二三四五六日'
    s1 = dt.strftime('%y%m%d') + f'周{weektag[dt.weekday()]}'  # '180827周一'

    # 2 获取后端数值标签
    d2 = int(datetime.now().strftime('%H%M%S'))

    # 3 获取目录下文件，并迭代确保生成一个不重名文件
    ls = os.listdir(fd)
    files = set(map(lambda x: os.path.basename(os.path.splitext(x)[0]), ls))  # 收集basename

    while s1 + str(d2) in files:
        d2 += 1

    return s1 + str(d2)


def myoswalk(root, filter_rule=None, recur=True):
    """
    :param root: 根目录
    :param filter_rule:
        字符串
            以点.开头的，统一认为是进行后缀格式识别
        其他字符串类型会认为是一个正则规则，只要相对root的全名能search到规则即认为匹配
            可以将中文问号用于匹配任意汉字
        也可以输入自定义函数： 输入参数是相对root目录下的文件全名
    :param recur: 是否进行子文件夹递归
    :return:
    """
    if isinstance(filter_rule, str):
        filter_rule = gen_file_filter(filter_rule)

    # prefix_len = len(root)  # 计算出前缀长度
    for dirpath, dirnames, filenames in os.walk(root):
        # relative_root = dirpath[prefix_len+1:]  # 我想返回相对路径，但是好像不太规范会对很多东西造成麻烦
        #  过滤掉特殊目录
        for t in ('.git', '$RECYCLE.BIN', '__pycache__', 'temp', 'Old', 'old'):
            try:
                del dirnames[dirnames.index(t)]
            except ValueError:
                pass
        # 去掉备份文件
        dirnames = list(filter(lambda x: not File(x).backup_time and '-冲突-' not in x, dirnames))
        filenames = list(filter(lambda x: not File(x).backup_time and '-冲突-' not in x, filenames))

        # 调用特殊过滤规则
        if filter_rule:
            dirnames = list(filter(lambda x: filter_rule(f'{dirpath}\\{x}'), dirnames))
            filenames = list(filter(lambda x: filter_rule(f'{dirpath}\\{x}'), filenames))

        # 如果该文件夹下已经没有文件，不返回该目录
        if not (filenames or dirnames):
            continue

        # 返回生成结果
        yield dirpath, dirnames, filenames

        if not recur:  # 不进行递归
            break


def mygetfiles(root, filter_rule=None, recur=True):
    r""" 对myoswalk进一步封装，返回所有匹配的文件
    会递归查找所有子文件

    可以这样遍历一个目录下的所有文件：
    for f in mygetfiles(r'C:\pycode\code4101py', r'.py'):
        print(f)
    这个函数已经自动过滤掉备份文件了
    筛选规则除了“.+后缀”，还可以写正则匹配

    参数含义详见myoswalk
    """
    for root, _, files in myoswalk(root, filter_rule, recur):
        for f in files:
            yield root + '\\' + f


def __6_high():
    """ 一些高级的路径功能 """


class DirsFileFinder:
    """ 多目录里的文件检索类 """

    def __init__(self, *dirs):
        """ 支持按优先级输入多个目录dirs，会对这些目录里的文件进行统一检索 """
        self.names = defaultdict(list)
        self.stems = defaultdict(list)

        for d in dirs:
            self.add_dir(d)

    def add_dir(self, p, cvt_name_func=None):
        """ 添加备用检索目录
        当前面的目录找不到匹配项的时候，会使用备用目录的文件
        备用目录可以一直添加，有多个，优先级逐渐降低

        :param cvt_name_func: 对名称做个转换再匹配
        """
        files = list(XlPath(p).rglob_files())
        if cvt_name_func:
            for f in files:
                self.names[cvt_name_func(f.name)].append(f)
                self.stems[cvt_name_func(f.stem)].append(f)
        else:
            for f in files:
                self.names[f.name].append(f)
                self.stems[f.stem].append(f)

    def find_name(self, name):
        """ 返回第一个匹配的结果 """
        files = self.find_names(name)
        if files:
            return files[0]

    def find_names(self, name):
        """ 返回所有匹配的结果 """
        return self.names[name]

    def find_stem(self, stem):
        files = self.find_stems(stem)
        if files:
            return files[0]

    def find_stems(self, stem):
        return self.stems[stem]

    def find_prefix_name(self, prefix_name, suffix=None):
        files = self.find_prefix_names(prefix_name, suffix=suffix)
        if files:
            return files[0]

    def find_prefix_names(self, prefix_name, suffix=None):
        """ name中前缀为prefix_name """
        filess = [files for name, files in self.names.items() if name.startswith(prefix_name)]
        # 将嵌套的list展平
        files = [file for files in filess for file in files]
        if suffix:
            files = [file for file in files if file.suffix == suffix]
        return files


class TwinDirs:
    def __init__(self, src_dir, dst_dir):
        """ 一对'孪生'目录，一般是有一个src_dir，还有个同结构的dst_dir。
        但dst_dir往往并不存在，是准备从src_dir处理过来的。
        """
        self.src_dir = XlPath(src_dir)
        self.dst_dir = XlPath(dst_dir)
        self.src_dir_finder = None

    def reset_dst_dir(self):
        """ 重置目标目录 """
        self.dst_dir.delete()
        self.dst_dir.mkdir()

    def copy_file(self, src_file, if_exists=None):
        """ 从src复制一个文件到dst里

        :param XlPath src_file: 原文件位置（其实输入目录类型也是可以的，内部实现逻辑一致的）
        """
        src_file = XlPath(src_file)
        _src_file = src_file.relpath(self.src_dir)
        dst_file = self.dst_dir / _src_file
        dst_dir = dst_file.parent
        dst_dir.mkdir(exist_ok=True, parents=True)  # 确保目录结构存在
        src_file.copy(dst_file, if_exists=if_exists)
        return dst_file

    def copy_ext_file(self, ext_file, if_exists=None):
        """ 和copy_file区别，这里输入的ext_file是来自其他目录的文件
        但是要在src_file里找到位置，按照结构复制到dst_dir
        """
        # 1 找到原始文件位置
        ext_file = XlPath(ext_file)
        if self.src_dir_finder is None:
            self.src_dir_finder = DirsFileFinder(self.src_dir)

        src_files = self.src_dir_finder.find_names(ext_file.name)
        assert len(src_files) < 2, '出现多种匹配可能性，请检查目录'

        if len(src_files) == 0:
            src_files = self.src_dir_finder.find_stems(ext_file.stem)
            assert len(src_files), '没有找到可匹配的文件'

        src_file = src_files[0]

        # 2 复制文件
        _src_file = src_file.relpath(self.src_dir)
        dst_file = self.dst_dir / _src_file
        dst_dir = dst_file.parent
        dst_dir.mkdir(exist_ok=True, parents=True)  # 确保目录结构存在
        ext_file.copy(dst_file, if_exists=if_exists)
        print(ext_file, '->', dst_file)
        return dst_file

    def copy_dir_structure(self):
        """ 复制目录结构 """
        self.src_dir.copy_dir_structure(self.dst_dir)


class BatchFileRenamer:
    """ 对一批数据，按照某种规则判重、重命名去重
    一般是对stem重命名后，确保数据随意混合后名称也不会有出现重复
    """

    def __init__(self, _dir=None):
        """
        :param _dir: 输入待处理的第一个目录
        """
        # 所有待处理的文件
        self.files = []
        if _dir is not None:
            self.add_dir(_dir)

    def add_dir(self, _dir):
        """ 添加一个目录下的所有文件
        如果有比较零散的文件待处理，可以直接操作self.files

        todo 目录的重命名？如果引入目录的重命名，算法会复杂非常多的，这个暂不考虑。
        """
        for f in XlPath(_dir).rglob_files():
            self.files.append(f)

    def get_key(self, file):
        """ 计算一个文件的重复标识，不同文件之间的判重依据 """
        # 默认的key，规则会比较严，stem不重复，大小写不重复
        return file.stem.lower()

    def get_new_name(self, file, exists_keys=None):
        """ 输入的f必须是已经确定要进行重命名的文件

        对于windows来说
            是先假定在一个组中，增加编号
            如果编号的文件已经存在，则换一个新的命名组

        对我这里来说，就不搞这么复杂了，就是无脑加编号就行
        """

        def add_stem(m):
            a = int(m.group(1)) + 1
            return f'({a})'

        stem = file.stem
        while True:
            m = re.search(r'\(\d+\)$', stem)
            if m:  # 已经有目标范式的编号，继续累加
                stem = re.sub(r'\((\d+)\)$', add_stem, stem)
            else:  # 还没有编号的，直接从'2'开始编号
                stem += ' (2)'

            f2 = file.with_stem(stem.strip())  # 忽略最后空白，这个很容易出问题
            k2 = self.get_key(f2)
            if k2 not in exists_keys:  # 如果新的命名不会跟旧有文件有任何重复，循环就可以终止了
                exists_keys.add(k2)
                return f2

    def rename_files(self, print_mode=False, exists_keys=None):
        """ 对文件进行批量重命名

        这里分两轮运行，是有非常深的用意的，避免前面重命名的时候，没有见到后面可能会出现的重名文件
        """
        # 1 先遍历一遍文件，确认哪些文件是确定要重命名的
        exists_keys = exists_keys or set()
        repeat_name_files = []  # 确认要进行重命名的文件
        for f in self.files:
            k = self.get_key(f)
            if k not in exists_keys:
                exists_keys.add(k)
            else:
                repeat_name_files.append(f)

        # 2 对需要重命名的文件进行操作
        cnt = 0
        for f in repeat_name_files:
            cnt += 1
            f2 = self.get_new_name(f, exists_keys)
            if print_mode:
                print(cnt, f.as_posix(), '-->', f2.name)
            f.rename2(f2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2022/06/09 16:26

"""
针对PostgreSQL封装的工具
"""

import os
import sys

from pyxllib.prog.pupil import check_install_package
from pyxllib.file.specialist import XlPath, ensure_localdir

# 使用binary安装，好像就不需要自己配置依赖了
# if sys.platform == 'win32':
#     # windows系统中pg的默认安装位置
#     # https://www.yuque.com/xlpr/pyxllib/install_psycopg
#     pgdll_dir = XlPath('C:/Program Files/PostgreSQL/14/bin')
#     if not pgdll_dir.is_dir():  # 换一个自定义目录继续检查
#         pgdll_dir = XlPath.userdir() / '.xlpr/pgdll'
#     if not pgdll_dir.is_dir():
#         print('没有PostgreSQL对应的dll文件，自动下载：')
#         ensure_localdir(pgdll_dir, 'https://xmutpriu.com/download/pgdll_v14.zip')
#     os.environ['PATH'] += ';' + pgdll_dir.as_posix()

check_install_package('psycopg', 'psycopg[binary]')

import io
from collections import Counter
import json
import json
import textwrap
import datetime
import re

from tqdm import tqdm

import psycopg
import psycopg.rows

from pyxllib.prog.newbie import round_int, human_readable_number
from pyxllib.prog.pupil import utc_now, utc_timestamp, is_valid_identifier
from pyxllib.prog.specialist import XlOsEnv
from pyxllib.algo.pupil import ValuesStat2
from pyxllib.file.specialist import get_etag, StreamJsonlWriter
from pyxllib.data.sqlite import SqlBase, SqlBuilder


class Connection(psycopg.Connection, SqlBase):

    def __init__(self, *args, **kwargs):
        psycopg.Connection.__init__(self, *args, **kwargs)
        SqlBase.__init__(self, *args, **kwargs)

    def __del__(self):
        self.close()

    def __1_库(self):
        pass

    def get_db_activities(self, datname=None):
        """
        检索当前数据库的活动信息。

        :param datname: 这个字段理论上应该要能自动检测出来才对，但这会急着没空钻研，先手动输入吧
        """
        sql = SqlBuilder('pg_stat_activity')
        sql.select('pid', 'datname', 'usename', 'state', 'query', 'age(now(), query_start) AS "query_age"')
        sql.where("state = 'active'")
        if datname:
            sql.where(f"datname = '{datname}'")
        return self.exec2dict(sql.build_select()).fetchall()

    def __2_表格(self):
        pass

    def get_table_names(self):
        # cmd = "SELECT table_name FROM information_schema.tables WHERE table_schema='public' AND table_type='BASE TABLE'"
        cmd = "SELECT tablename FROM pg_tables WHERE schemaname='public'"
        return [x[0] for x in self.execute(cmd)]

    def has_table(self, table_name, schemaname='public'):
        query = ["SELECT EXISTS (SELECT FROM pg_tables ",
                 f"WHERE schemaname='{schemaname}' AND tablename='{table_name}')"]
        res = self.execute(' '.join(query))
        return res.fetchone()[0]

    def get_column_names(self, table_name):
        """ 【查】表格有哪些字段
        """
        cmd = f"SELECT column_name FROM information_schema.columns WHERE table_name='{table_name}'"
        return self.exec2col(cmd)

    def ensure_column(self, table_name, col_name, *args, comment=None, **kwargs):
        super(Connection, self).ensure_column(table_name, col_name, *args, **kwargs)
        if comment:
            self.set_column_comment(table_name, col_name, comment)

    def set_column_comment(self, table_name, col_name, comment):
        # 这里comment按%s填充会报错。好像是psycopg库的问题，这种情况因为没有表格字段参照类型，会不知道要把comment转成字符串。
        # 然后使用py的f字符串填充的话，要避免comment有引号等特殊字符，就用了特殊的转义标记$zyzf$
        # 暂时找不到更优雅的方式~~ 而且这个问题，可能以后其他特殊的sql语句也会遇到。。。
        self.execute(f"COMMENT ON COLUMN {table_name}.{col_name} IS $zyzf${comment}$zyzf$")
        self.commit()

    def reset_table_item_id(self, table_name, item_id_name=None, counter_name=None):
        """ 重置表格的数据的id值，也重置计数器

        :param item_id_name: 表格的自增id字段名，有一套我自己风格的自动推算算法
        :param counter_name: 计数器字段名，如果有的话，也会重置
        """
        # 1 重置数据的编号
        if item_id_name is None:
            m = re.match(r'(.+?)(_table)?$', table_name)
            item_id_name = m.group(1) + '_id'

        sql = f"""WITH cte AS (
    SELECT {item_id_name}, ROW_NUMBER() OVER (ORDER BY {item_id_name}) AS new_{item_id_name}
    FROM {table_name}
)
UPDATE {table_name}
SET {item_id_name} = cte.new_{item_id_name}
FROM cte
WHERE {table_name}.{item_id_name} = cte.{item_id_name}"""
        self.execute(sql)  # todo 这种sql写法好像偶尔有bug会出问题
        self.commit()

        # 2 重置计数器
        if counter_name is None:
            counter_name = f'{table_name}_{item_id_name}_seq'

        # 找到目前最大的id值
        max_id = self.exec2one(f'SELECT MAX({item_id_name}) FROM {table_name}')
        # self.execute(f'ALTER SEQUENCE {counter_name} RESTART WITH {max_id + 1}')
        # 检查序列是否存在，如果不存在则创建序列，然后重置序列
        # 检查序列是否存在，如果不存在则创建序列，并将其关联到指定的表和字段
        self.execute(f"""
DO $$
BEGIN
    IF NOT EXISTS (SELECT FROM pg_class WHERE relkind = 'S' AND relname = '{counter_name}') THEN
        EXECUTE format('CREATE SEQUENCE %I', '{counter_name}');
        EXECUTE format('ALTER TABLE %I ALTER COLUMN %I SET DEFAULT nextval(''%I'')', '{table_name}', '{item_id_name}', '{counter_name}');
    END IF;
    EXECUTE format('ALTER SEQUENCE %I RESTART WITH %s', '{counter_name}', {max_id + 1});
END $$;
""")
        self.commit()

    def __3_execute(self):
        pass

    def exec_nametuple(self, *args, **kwargs):
        cur = self.cursor(row_factory=psycopg.rows.namedtuple_row)
        data = cur.execute(*args, **kwargs)
        # cur.close()
        return data

    def exec2dict(self, *args, **kwargs):
        cur = self.cursor(row_factory=psycopg.rows.dict_row)
        data = cur.execute(*args, **kwargs)
        # cur.close()
        return data

    def exec2dict_batch(self, sql, batch_size=1000, use_offset=None, **kwargs):
        """ 分批返回数据的版本

        :param use_offset: 是否使用offset分页，会根据sql中是否含有where自动判断，但有时候最好明确指定以防错误
            如果外部sql每次操作，会改变数据库的情况，导致sql的where规则虽然没变，但是数据本身发生变化，则offset应该要关闭
                每次取对应的满足条件的数据即可
                这种情况，也需要本函数内部主动执行commit_all的
            否则，只是一种遍历查询，没有where或者where获取的数据情况是不会变化的，则要使用offset
        :return:
            第1个值，是一个迭代器，看起来仍然能一条一条返回，实际后台是按照batch_size打包获取的
            第2个值，是数据总数
        """
        if not isinstance(sql, SqlBuilder):
            raise ValueError('暂时只能搭配SQLBuilder使用')

        if use_offset is None:
            use_offset = not sql._where

        num = self.exec2one(sql.build_count())
        offset = 0

        def yield_row():
            nonlocal offset
            while True:
                sql2 = sql.copy()
                if not use_offset:  # 如果不使用offset，那么缓存的sql操作需要全部提交，确保数据都更新后，再提取数据
                    self.commit_all()
                sql2.limit(batch_size, offset)
                rows = self.exec2dict(sql2.build_select(), **kwargs).fetchall()
                if use_offset:
                    offset += len(rows)
                if not rows:
                    break
                yield from rows

        return yield_row(), num

    exec_dict = exec2dict

    def __4_数据类型(self):
        pass

    @classmethod
    def cvt_type(cls, val):
        if isinstance(val, (dict, list)):
            val = json.dumps(val, ensure_ascii=False, default=str)
        # 注意list数组类型读、写都会自动适配py
        return val

    @classmethod
    def autotype(cls, val):
        if isinstance(val, str):
            return 'text'
        elif isinstance(val, int):
            return 'int4'  # int2、int8
        elif isinstance(val, bool):
            return 'boolean'
        elif isinstance(val, float):
            return 'float4'
        elif isinstance(val, (dict, list)):
            return 'jsonb'
        elif isinstance(val, datetime.datetime):
            return 'timestamp'
        else:  # 其他list等类型，可以用json.dumps或str转文本存储
            return 'text'

    def __5_增删改查(self):
        pass

    def insert_row(self, table_name, cols, *, on_conflict='DO NOTHING', commit=False):
        """ 【增】插入新数据

        :param dict cols: 用字典表示的要插入的值
        :param on_conflict: 如果已存在的处理策略
            DO NOTHING，跳过不处理
            REPLACE，这是个特殊标记，会转为主键冲突也仍然更新所有值
            (id) DO UPDATE SET host_name=EXCLUDED.host_name, nick_name='abc'
                也可以写复杂的处理算法规则，详见 http://postgres.cn/docs/12/sql-insert.html
                比如这里是插入的id重复的话，就把host_name替换掉，还可以指定nick_name替换为'abc'
                注意前面的(id)是必须要输入的

        注意：有个常见需求，是想插入后返回对应的id，但是这样就需要知道这张表自增的id字段名
            以及还是很难获得插入后的id值，可以默认刚插入的id是最大的，但是这样并不安全，有风险
            建议还是外部自己先计算全表最大的id值，自己实现自增，就能知道插入的这条数据的id了
        """
        ks = ','.join(cols.keys())
        vs = ','.join(['%s'] * (len(cols.keys())))
        query = f'INSERT INTO {table_name}({ks}) VALUES ({vs})'
        params = self.cvt_types(cols.values())

        if on_conflict == 'REPLACE':
            on_conflict = f'ON CONFLICT ON CONSTRAINT {table_name}_pkey DO UPDATE SET ' + \
                          ','.join([f'{k}=EXCLUDED.{k}' for k in cols.keys()])
        else:
            on_conflict = f'ON CONFLICT {on_conflict}'
        query += f' {on_conflict}'

        self.commit_base(commit, query, params)

    def insert_rows(self, table_name, keys, ls, *, on_conflict='DO NOTHING', commit=False):
        """ 【增】插入新数据

        :param str keys: 要插入的字段名，一个字符串，逗号,隔开属性值
        :param list[list] ls: n行m列的数组

        >> con.insert_rows('hosts2', 'id,host_name,nick_name', [[1, 'test5', 'dcba'], [11, 'test', 'aabb']])
        """
        n, m = len(ls), len(keys.split(','))
        vs = ','.join(['%s'] * m)
        query = f'INSERT INTO {table_name}({keys}) VALUES ' + ','.join([f'({vs})'] * n)
        params = []
        for cols in ls:
            params += self.cvt_types(cols)

        if on_conflict == 'REPLACE':
            on_conflict = f'ON CONFLICT ON CONSTRAINT {table_name}_pkey DO UPDATE SET ' + \
                          ','.join([f'{k.strip()}=EXCLUDED.{k.strip()}' for k in keys.split(',')])
        else:
            on_conflict = f'ON CONFLICT {on_conflict}'
        query += f' {on_conflict}'

        self.commit_base(commit, query, params)

    def __6_高级统计(self):
        pass

    def get_column_valuesstat(self, table_name, column, filter_condition=None,
                              percentile_count=5,
                              by_data=False, data_type=None):
        """ 获得指定表格的某个字段的统计特征ValuesStat2对象

        :param table_name: 表名
        :param column: 用于计算统计数据的字段名
            不一定是标准的字段名
        :param percentile_count: 分位数的数量，例如 3 表示只计算中位数
        :param by_data: 是否获得原始数据
            默认只获得统计特征，不获得原始数据
        """

        def init_from_db_data():
            sql = SqlBuilder(table_name)
            if filter_condition:
                sql.where(filter_condition)
            values = self.exec2col(sql.build_select(column))
            if data_type == 'numeric':
                values = [x and float(x) for x in values]
            vs = ValuesStat2(raw_values=values, data_type=data_type)

            if data_type == 'text' and is_valid_identifier(column):
                vs0 = self.get_column_valuesstat(table_name, column, filter_condition=filter_condition,
                                                 percentile_count=percentile_count, by_data=False)
                vs.n = vs0.n
                vs.dist = vs0.dist

            return vs

        def init_from_db():
            # 1 构建基础的 SQL 查询
            sql = SqlBuilder(table_name)
            sql.select("COUNT(*) AS total_count")
            sql.select(f"COUNT({column}) AS non_null_count")
            sql.select(f"MIN({column}) AS min_value")
            sql.select(f"MAX({column}) AS max_value")
            if data_type and 'timestamp' in data_type:
                percentile_type = 'PERCENTILE_DISC'
                # todo 其实时间类也可以"泛化"一种平均值、标准差算法的，这需要获取全量数据，然后自己计算
            elif data_type == 'text':
                percentile_type = 'PERCENTILE_DISC'
            else:  # 默认是正常的数值类型
                sql.select(f"SUM({column}) AS total_sum")
                sql.select(f"AVG({column}) AS average")
                sql.select(f"STDDEV({column}) AS standard_deviation")
                percentile_type = 'PERCENTILE_CONT'

            percentiles = []
            # 根据分位点的数量动态添加分位数计算
            if percentile_count > 2:
                step = 1 / (percentile_count - 1)
                percentiles = [(i * step) for i in range(1, percentile_count - 1)]
                for p in percentiles:
                    sql.select(f"{percentile_type}({p:.2f}) WITHIN GROUP (ORDER BY {column}) "
                               f"AS percentile_{int(p * 100)}")

            if filter_condition:
                sql.where(filter_condition)

            row = self.exec2dict(sql.build_select()).fetchone()

            # 2 统计展示
            x = ValuesStat2(data_type=data_type)
            x.raw_n = row['total_count']
            x.n = row['non_null_count']
            if not x.n:
                return x

            x.sum = row.get('total_sum', None)
            x.mean = row.get('average', None)
            x.std = row.get('standard_deviation', None)

            # 如果计算了分位数，填充相应属性
            x.dist = [row['min_value']] + [row[f"percentile_{int(p * 100)}"] for p in percentiles] + [row['max_value']]
            if data_type == 'numeric':
                x.dist = [float(x) for x in x.dist]

            return x

        data_type = data_type or self.get_column_data_type(table_name, column)

        # 如果不是标准的列名，强制获取数据
        if not is_valid_identifier(column):
            by_data = True

        if by_data:
            vs = init_from_db_data()
        else:
            vs = init_from_db()

        return vs

    def export_jsonl(self, file_path, table_name, key_col=None, batch_size=1000, print_mode=0):
        """ 将某个表导出为本地jsonl文件

        :param str|SqlBuilder table_name: 表名
            支持传入SqlBuilder对象，这样可以更灵活的控制导出的数据规则
        :param file_path: 导出的文件路径
        :param batch_size: 每次读取的行数和保存的行数
        :param key_col: 作为主键的列名，如果有的话，会自动去重
            强烈推荐要设置
            实际不一定要用主键，只要是有顺序值的列就行

        todo 暴力最简单的版本不难写，我纠结的是缓存机制，还有bytes类型数据会有点大等问题
            还需要先支持一个通用的缓存写文件功能
        """
        # 1 sql
        if isinstance(table_name, str):
            sql = SqlBuilder(table_name)
            sql.select('*')
        else:
            sql = table_name
            m = re.search(r'FROM (\w+)', sql.build_select())
            table_name = m.group(1) if m else 'table'
        assert isinstance(sql, SqlBuilder)

        file_path = XlPath(file_path)
        if key_col:
            sql.order_by(key_col)
            if file_path.is_file():
                # 读取现有数据，找出主键最大值
                data = file_path.read_jsonl(batch_size=1000)
                if data:
                    max_val = max([x[key_col] for x in data]) if data else None
                    if max_val is not None:
                        sql.where(f'{key_col} > {max_val}')

        # 2 获取数据
        file = StreamJsonlWriter(file_path, batch_size=batch_size)  # 流式存储
        rows, total = self.exec2dict_batch(sql, batch_size=batch_size, use_offset=True)
        for row in tqdm(rows, total=total, desc=f'从{table_name}表导出数据', disable=not print_mode):
            file.append_line(row)
        file.flush()

    def check_db_tables_size(self, db_name=None):
        """ 查看指定数据下所有表格的大小 """
        from datetime import datetime
        import pandas as pd

        if db_name is None:
            # 使用sql获取当前self所在数据库
            db_name = self.exec2one("SELECT current_database()")

        data = []
        tables = self.exec2col("SELECT table_name FROM information_schema.tables WHERE table_schema='public'")
        for table_name in tables:
            row = {
                'database': db_name,
                'table_name': table_name,
            }
            sz = self.exec2one(f"SELECT pg_total_relation_size('public.{table_name}')")
            if not sz:
                continue
            lines = self.exec2one(f"SELECT COUNT(*) FROM {table_name}")
            row['size'], row['lines'] = sz, lines
            row['readable_size'] = human_readable_number(sz, 'KB')
            row['perline_size'] = human_readable_number(sz / lines, 'KB') if lines else -1
            row['update_time'] = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            data.append(row)

        df = pd.DataFrame.from_dict(data)
        if len(df):
            df.sort_values(['size'], ascending=False, inplace=True)
        df.reset_index(drop=True, inplace=True)
        return df

    def check_multi_db_size(self, db_list):
        """ 这个功能一般要用postgres账号，才有权限处理所有数据库 """
        from datetime import datetime
        import pandas as pd

        data = []
        for db in db_list:
            row = {
                'name': db,
            }
            sz = self.exec2one(f"SELECT pg_database_size('{db}')")
            row['size'] = sz
            row['readable_size'] = human_readable_number(sz, 'KB')
            row['update_time'] = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

            data.append(row)

        df = pd.DataFrame.from_dict(data)
        df.sort_values(['size'], ascending=False, inplace=True)
        df.reset_index(drop=True, inplace=True)
        return df


"""
【关于为什么XlprDb要和pglib合一个文件】
好处
1、代码集中管理，方便开发效率

坏处 - 对治
1、导入代码多，效率低
    - 导入代码其实很快的，不差这一点点
2、仅用简单功能的时候，会引入多余复杂功能
    - 如果单拉一个xlprdb.py文件，依然不能头部import复杂库，因为XlprDb也存在需要精简使用的场景
        既然都要有精简的部分，干脆都把代码弄到一起得了
        然后复杂包都在成员函数里单独import
"""


class XlprDb(Connection):
    """ xlpr统一集中管理的一个数据库

    为了一些基础的数据库功能操作干净，尽量不要在全局导入过于复杂的包，可以在每个特定功能里导特定包
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.seckey = ''

    @classmethod
    def set_conninfo(cls, conninfo, seckey=''):
        """ 提前将登录信息加密写到环境变量中，这样代码中就没有明文的账号密码信息

        :param conninfo:
        :param seckey: 如果要获得数据库里较重要的密码等信息，需要配置key值，否则默认可以不设

        使用后，需要重启IDE重新加载环境变量
        并且本句明文代码需要删除
        """
        # TODO 目前只设一个账号，后续可以扩展支持多个账号指定配置
        # conninfo = 'postgresql://postgres:yourpassword@172.16.170.110/xlpr'
        return XlOsEnv.persist_set('XlprDbAccount', {'conninfo': conninfo, 'seckey': seckey}, encoding=True)

    @classmethod
    def connect(cls, conninfo='', seckey='', *,
                autocommit=False, row_factory=None, context=None, **kwargs) -> 'XlprDb':
        """ 因为要标记 -> XlprDb，IDE才会识别出类别，有自动补全功能
        但在类中写@classmethod，无法标记 -> XlprDb，所以就放外面单独写一个方法了
        """
        d = XlOsEnv.get('XlprDbAccount', decoding=True)
        if conninfo == '':
            conninfo = d['conninfo']
        if seckey == '' and isinstance(d, dict) and 'seckey' in d:
            seckey = d['seckey']
        # 注意这里获取的是XlprDb类型
        con = super(XlprDb, cls).connect(conninfo, autocommit=autocommit, row_factory=row_factory, context=context,
                                         **kwargs)
        con.seckey = seckey
        return con

    @classmethod
    def connect2(cls, name, passwd, database=None, ip_list=None, **kwargs):
        """ 简化的登录方式，并且自带优先尝试局域网连接，再公网连接

        一般用在jupyter等对启动速度没有太高要求的场合，因为不断尝试localhost等需要耗费不少时间
        """
        if database is None:
            database = name

        xldb = None
        if ip_list is None:
            ip_list = ['localhost', '172.16.170.136', 'xmutpriu.com']
        for ip in ip_list:
            try:
                xldb = cls.connect(f'postgresql://{name}:{passwd}@{ip}/{database}', **kwargs)
                break
            except psycopg.OperationalError:
                pass

        return xldb

    def __1_hosts相关数据表操作(self):
        pass

    def update_host(self, host_name, accounts=None, **kwargs):
        """ 更新一台服务器的信息

        :param dict accounts: 账号信息，记得要列全，比如
            {'root': '123456', 'chenkunze': '654321'}
        """
        if not self.execute('SELECT EXISTS (SELECT FROM hosts WHERE host_name=%s)', (host_name,)).fetchone()[0]:
            self.insert_row('hosts', {'host_name': host_name})
        if kwargs:
            self.update_row('hosts', kwargs, {'host_name': host_name})
        if accounts:
            self.execute('UPDATE hosts SET accounts=pgp_sym_encrypt(%s, %s) WHERE host_name=%s',
                         (json.dumps(accounts, ensure_ascii=False), self.seckey, host_name))
        self.commit()

    def set_host_account(self, host_name, user_name, passwd):
        """ 设置某台服务器的一个账号密码

        >> xldb.set_host_account('titan2', 'chenkunze', '123456')

        """
        # 读取旧的账号密码字典数据
        d = self.execute("SELECT pgp_sym_decrypt(accounts, %s)::jsonb FROM hosts WHERE host_name=%s",
                         (self.seckey, host_name)).fetchone()[0]
        # 修改字典数据
        d[user_name] = str(passwd)
        # 将新的字典数据写回数据库
        self.execute('UPDATE hosts SET accounts=pgp_sym_encrypt(%s, %s) WHERE host_name=%s',
                     (json.dumps(d, ensure_ascii=False), self.seckey, host_name))
        self.commit()

    def login_ssh(self, host_name, user_name, map_path=None, **kwargs) -> 'XlSSHClient':
        """ 通过数据库里的服务器数据记录，直接登录服务器 """
        from pyxllib.ext.unixlib import XlSSHClient

        if host_name.startswith('g_'):
            host_ip = self.execute("SELECT host_ip FROM hosts WHERE host_name='xlpr0'").fetchone()[0]
            pw, port = self.execute('SELECT (pgp_sym_decrypt(accounts, %s)::jsonb)[%s]::text, frpc_port'
                                    ' FROM hosts WHERE host_name=%s',
                                    (self.seckey, user_name, host_name[2:])).fetchone()
        else:
            port = 22
            host_ip, pw = self.execute('SELECT host_ip, (pgp_sym_decrypt(accounts, %s)::jsonb)[%s]::text'
                                       ' FROM hosts WHERE host_name=%s',
                                       (self.seckey, user_name, host_name)).fetchone()

        if map_path is None:
            if sys.platform == 'win32':
                map_path = {'C:/': '/'}
            else:
                map_path = {'/': '/'}

        return XlSSHClient(host_ip, user_name, pw[1:-1], port=port, map_path=map_path, **kwargs)

    def __2_xlapi相关数据表操作(self):
        """
        files，存储二进制文件的表
            etag，文件、二进制对应的etag值
            meta，可以存储不同数据类型一些特殊的属性，比如图片可以存储高、宽
                但因为用户使用中，为了提高速度，以及减少PIL等依赖，执行中不做计算
                可以其他途径使用定期脚本自动处理
        xlapi，底层api调用的记录统计
            input，所有输入的参数
                mode，使用的算法接口
                etag，涉及到大文件数据的，以打包后的文件的etag做记录
            output，运行结果
            elapse_ms，调用函数的用时，不含后处理转换格式的时间
        xlserver，服务端记录的使用情况
        """
        pass

    def insert_row2files(self, buffer, *, etag=None, **kwargs):
        """

        为了运算效率考虑，除了etag需要用于去重，是必填字段
        其他字段默认先不填充计算
        """
        # 1 已经有的不重复存储
        if etag is None:
            etag = get_etag(buffer)

        res = self.execute('SELECT etag FROM files WHERE etag=%s', (etag,)).fetchone()
        if res:  # 已经做过记录的，不再重复记录
            return

        # 2 没有的图，做个备份
        kwargs['etag'] = etag
        kwargs['data'] = buffer
        kwargs['fsize_kb'] = round_int(len(buffer) / 1024)

        self.insert_row('files', kwargs)
        self.commit()

    def get_xlapi_record(self, **input):
        res = self.execute('SELECT id, output FROM xlapi WHERE input=%s', (self.cvt_type(input),)).fetchone()
        if res:
            _id, output = res
            output['xlapi_id'] = _id
            return output

    def insert_row2xlapi(self, input, output, elapse_ms, *, on_conflict='(input) DO NOTHING'):
        """ 往数据库记录当前操作内容

        :return: 这个函数比较特殊，需要返回插入的条目的id值
        """
        if on_conflict == 'REPLACE':
            on_conflict = "(input) DO UPDATE " \
                          "SET output=EXCLUDED.output, elapse_ms=EXCLUDED.elapse_ms, update_time=EXCLUDED.update_time"

        input = json.dumps(input, ensure_ascii=False)
        self.insert_row('xlapi', {'input': input, 'output': output,
                                  'elapse_ms': elapse_ms, 'update_time': utc_timestamp(8)},
                        on_conflict=on_conflict)
        self.commit()
        return self.execute('SELECT id FROM xlapi WHERE input=%s', (input,)).fetchone()[0]

    def insert_row2xlserver(self, request, xlapi_id=0, **kwargs):
        kw = {'remote_addr': request.headers.get('X-Real-IP', request.remote_addr),
              'route': '/'.join(request.base_url.split('/')[3:]),
              'update_time': utc_timestamp(8),
              'xlapi_id': xlapi_id}
        kw.update(kwargs)
        print(kw)  # 监控谁在用api
        self.insert_row('xlserver', kw, commit=True)

    def __3_host_trace相关可视化(self):
        """ TODO dbview 改名 host_trace """
        pass

    def __dbtool(self):
        pass

    def record_host_usage(self, cpu=True, gpu=True, disk=False):
        """ 记录服务器各种状况，存储到PG数据库

        TODO 并行处理
        TODO 功能还可以增加：gpu显卡温度、硬盘读写速率检查、网络上传下载带宽
        """
        # 1 服务器列表
        host_names = self.exec2col('SELECT host_name FROM hosts WHERE id > 1 ORDER BY id')
        host_cpu_gb = {h: v for h, v in self.execute('SELECT host_name, cpu_gb FROM hosts')}

        # 2 去所有服务器取使用情况
        for i, host_name in enumerate(host_names, start=1):
            print('-' * 20, i, host_name, '-' * 20)
            try:
                ssh = self.login_ssh(host_name, 'root', relogin=5, relogin_interval=0.2)
                status = {}
                if cpu:
                    data = ssh.check_cpu_usage(print_mode=True)
                    status['cpu'] = {k: v[0] for k, v in data.items()}
                    status['cpu_memory'] = {k: round(v[1] * host_cpu_gb[host_name] / 100, 2) for k, v in data.items()}
                if gpu:
                    status['gpu_memory'] = ssh.check_gpu_usage(print_mode=True)
                if disk:
                    # 检查磁盘空间会很慢，如果超时可以跳过。（设置超时6小时）
                    status['disk_memory'] = ssh.check_disk_usage(print_mode=True, timeout=60 * 60 * 6)
            except Exception as e:
                status = {'error': f'{str(type(e))[8:-2]}: {e}'}
                print(status)

            if status:
                self.insert_row('host_trace',
                                {'host_name': host_name, 'status': status, 'update_time': utc_timestamp(8)},
                                commit=True)
            print()

    def _get_host_trace_total(self, mode, title, yaxis_name, date_trunc, recent, host_attr):
        # CREATE INDEX ON gpu_trace (update_time);  -- update_time最好建个索引
        ls = self.execute(textwrap.dedent(f"""\
        WITH cte1 AS (  -- 筛选近期内的数据，并且时间做trunc处理
            SELECT host_name, (status)['{mode}'], date_trunc('{date_trunc}', update_time) ttime
            FROM host_trace WHERE update_time > %s AND (status ? '{mode}')
        ), cte2 AS (  -- 每个时间里每个服务器的多条记录取平均
            SELECT ttime, cte1.host_name, jsonb_div(jsonb_deep_sum(status), count(*)) status
            FROM cte1 GROUP BY ttime, cte1.host_name
        )  -- 接上每台服务器显存总值，并且分组得到每个时刻总情况
        SELECT ttime, {host_attr}, jsonb_deep_sum(status)
        FROM cte2 JOIN hosts ON cte2.host_name = hosts.host_name
        GROUP BY ttime ORDER BY ttime"""), ((utc_now(8) - recent).isoformat(timespec='seconds'),)).fetchall()
        return self._create_stack_chart(title, ls, yaxis_name=yaxis_name)

    def _get_host_trace_per_host(self, hostname, mode, title, yaxis_name, date_trunc, recent, host_attr):
        ls = self.execute(textwrap.dedent(f"""\
        WITH cte1 AS (
            SELECT (status)['{mode}'], date_trunc('{date_trunc}', update_time) ttime
            FROM host_trace WHERE host_name='{hostname}' AND update_time > %s AND (status ? '{mode}')
        ), cte2 AS (
            SELECT ttime, jsonb_div(jsonb_deep_sum(status), count(*)) status
            FROM cte1 GROUP BY ttime
        )
        SELECT ttime, {host_attr}, jsonb_deep_sum(status)
        FROM cte2 JOIN hosts ON hosts.host_name='{hostname}'
        GROUP BY ttime ORDER BY ttime"""), ((utc_now(8) - recent).isoformat(timespec='seconds'),)).fetchall()
        if not ls:  # 有的服务器可能数据是空的
            return '', 0
        return self._create_stack_chart(title, ls, yaxis_name=yaxis_name)

    def _create_stack_chart(self, title, ls, *, yaxis_name=''):
        """ 创建展示表

        :param title: 表格标题
        :param list ls: n*3，第1列是时间，第2列是总值，第3列是每个用户具体的数据
        """
        from pyecharts.charts import Line

        map_user_name = {}
        for ks, v in self.execute('SELECT account_names, name FROM users'):
            for k in ks:
                map_user_name[k] = v

        # 1 计算涉及的所有用户以及使用总量
        all_users_usaged = Counter()
        last_time = None
        for x in ls:
            hours = 0 if last_time is None else ((x[0] - last_time).total_seconds() / 3600)
            last_time = x[0]
            for k, v in x[2].items():
                if k == '_total':
                    continue
                all_users_usaged[map_user_name.get(k, k)] += v * hours

        # ls里的姓名也要跟着更新
        for i, x in enumerate(ls):
            ct = Counter()
            for k, v in x[2].items():
                ct[map_user_name.get(k, k)] += v
            ls[i] = (x[0], x[1], ct)

        # 2 转图表可视化
        def to_list(values):
            return [(x[0], v) for x, v in zip(ls, values)]

        def pretty_val(v):
            return round_int(v) if v > 100 else round(v, 2)

        try:
            chart = Line()
            chart.set_title(title)
            chart.options['xAxis'][0].update({'min': ls[0][0], 'type': 'time',
                                              # 'minInterval': 3600 * 1000 * 24,
                                              'name': '时间', 'nameGap': 50, 'nameLocation': 'middle'})
            chart.options['yAxis'][0].update({'name': yaxis_name, 'nameGap': 50, 'nameLocation': 'middle'})
            # 目前是比较暴力的方法调整排版，后续要研究是不是能更自动灵活些
            chart.options['legend'][0].update({'top': '6%', 'icon': 'pin'})
            chart.options['grid'] = [{'top': 55 + len(all_users_usaged) * 4, 'containLabel': True}]
            chart.options['tooltip'].opts.update({'axisPointer': {'type': 'cross'}, 'trigger': 'item'})

            chart.add_series(f'total{pretty_val(ls[0][1]):g}', to_list([x[1] for x in ls]), areaStyle={})
            for user, usaged in all_users_usaged.most_common():
                usaged = usaged / ((ls[-1][0] - ls[0][0]).total_seconds() / 3600 + 1e-9)
                chart.add_series(f'{user}{pretty_val(usaged):g}',
                                 to_list([x[2].get(user, 0) for x in ls]),
                                 areaStyle={}, stack='Total', emphasis={'focus': 'series'})

            return '<body>' + chart.render_embed() + '</body>', sum(all_users_usaged.values())
        except Exception as e:
            return str(e), 0

    #   cdx_edit
    def _get_database_trace_total(self, title, yaxis_name, date_trunc, recent, link_name):
        ls = self.execute(textwrap.dedent(f"""\
            WITH cte1 AS(
                SELECT link_name, jsonb_each(status::jsonb) AS db_data, date_trunc('{date_trunc}', update_time) ttime
                FROM database_trace WHERE update_time > %s AND link_name = '{link_name}'
            ), cte2 AS(
                SELECT ttime, link_name, (db_data).key AS table_name, ((db_data).value->> '_total')::bigint AS total
                FROM cte1
            )
            SELECT ttime, jsonb_object_agg(table_name, total) AS aggregated_json,SUM(total) as total
            FROM cte2
            GROUP BY ttime
            ORDER BY ttime"""), ((utc_now(8) - recent).isoformat(timespec='seconds'),)).fetchall()
        return self.database_create_stack_chart(title, ls, yaxis_name=yaxis_name)

    def _get_database_trace_per_host(self, db, title, yaxis_name, date_trunc, recent, link_name):
        ls = self.execute(textwrap.dedent(f"""\
            WITH cte1 AS (
                SELECT link_name, jsonb_each(status::jsonb) AS db_data, date_trunc('{date_trunc}', update_time) ttime
                FROM database_trace WHERE update_time > %s AND link_name = '{link_name}'
                ), cte2 AS (
                    SELECT ttime, link_name, (db_data).key AS table_name, (db_data).value AS size_text
                    FROM cte1
                ), cte3 AS (
                    SELECT ttime, table_name, each.key AS key, each.value AS value
                    FROM cte2, jsonb_each_text(size_text) AS each(key, value)
                )
                SELECT ttime, jsonb_object_agg(key,
                    CASE
                        WHEN key = '_total' THEN NULL
                        ELSE (value::jsonb ->> 'size')::bigint  -- Handle other keys as usual
                    END
                ) FILTER (WHERE key != '_total') AS aggregated_result,  -- 确保 _total 不在 aggregated_result 中
                MAX(CASE WHEN key = '_total' THEN value::bigint ELSE NULL END) AS total  -- 单独提取 _total 的值
                FROM cte3
                WHERE (key = '_total' OR value::jsonb ? 'size')  -- Ensure that '_total' is included
                AND table_name = '{db}'
                GROUP BY ttime
                ORDER BY ttime"""), ((utc_now(8) - recent).isoformat(timespec='seconds'),)).fetchall()
        return self.database_create_stack_chart(title, ls, yaxis_name=yaxis_name)

    def database_create_stack_chart(self, title, ls, *, yaxis_name=''):
        """ 创建展示表
    
        :param title: 表格标题
        :param list ls: n*3，第1列是时间，第3列是总值，第2列是每个用户具体的数据
        """
        from pyecharts.charts import Line
        all_database_usaged = Counter()
        last_time = None
        for x in ls:
            hours = 0 if last_time is None else ((x[0] - last_time).total_seconds() / 3600)
            last_time = x[0]
            for k, v in x[1].items():
                if k == '_total':
                    continue
                all_database_usaged[k] += v * hours

        for i, x in enumerate(ls):
            ct = Counter()
            for k, v in x[1].items():
                ct[k] += v
            ls[i] = (x[0], ct, int(x[2]))

        # 2 转图表可视化
        def to_list(values):
            return [(x[0], v) for x, v in zip(ls, values)]

        def pretty_val(v):
            return round_int(v) if v > 100 else round(v, 2)

        chart = Line()
        chart.set_title(title)
        chart.options['xAxis'][0].update({'min': ls[0][0], 'type': 'time',
                                          # 'minInterval': 3600 * 1000 * 24,
                                          'name': '时间', 'nameGap': 50, 'nameLocation': 'middle'})
        chart.options['yAxis'][0].update({'name': yaxis_name, 'nameGap': 50, 'nameLocation': 'middle'})
        # 目前是比较暴力的方法调整排版，后续要研究是不是能更自动灵活些
        chart.options['legend'][0].update({'top': '6%', 'icon': 'pin'})
        chart.options['grid'] = [{'top': 55 + len(all_database_usaged) * 4, 'containLabel': True}]
        chart.options['tooltip'].opts.update({'axisPointer': {'type': 'cross'}, 'trigger': 'item'})

        # chart.add_series(f'total {pretty_val(ls[0][2] / 1024 / 1024 / 1024):g}',
        #                 to_list([x[2] / 1024 / 1024 / 1024 for x in ls]), areaStyle={})
        for database, usaged in all_database_usaged.most_common():
            usaged = usaged / ((ls[-1][0] - ls[0][0]).total_seconds() / 3600 + 1e-9)
            chart.add_series(f'{database} {pretty_val(usaged / 1024 / 1024 / 1024):g}',
                             to_list([x[1].get(database, 0) / 1024 / 1024 / 1024 for x in ls]),
                             areaStyle={}, stack='Total', emphasis={'focus': 'series'})
        return '<body>' + chart.render_embed() + '</body>'

    def dbview_xldb1_memory(self, recent=datetime.timedelta(days=180), date_trunc='day'):
        from pyxllib.data.echarts import render_echart_html

        db_list = ['stdata', 'xlpr', 'st', 'ckz']
        args = ['数据库大小(GB)', date_trunc, recent, 'xldb1']
        htmltexts = []

        res = self._get_database_trace_total('xldb1数据库使用近况', *args)
        htmltexts.append(res)

        data_stats = []
        for idx, db in enumerate(db_list, start=1):
            data_stats.append(self._get_database_trace_per_host(db, f'{db}', *args))
        htmltexts += data_stats

        self.commit()
        h = render_echart_html('database_cdx', body='<br/>'.join(htmltexts))
        return h

    def dbview_xldb2_memory(self, recent=datetime.timedelta(days=180), date_trunc='day'):
        from pyxllib.data.echarts import render_echart_html

        db_list = ['ragdata', 'kq5034']
        args = ['数据库大小(GB)', date_trunc, recent, 'xldb2']
        htmltexts = []

        res = self._get_database_trace_total('xldb2数据库使用近况', *args)
        htmltexts.append(res)

        data_stats = []
        for idx, db in enumerate(db_list, start=1):
            data_stats.append(self._get_database_trace_per_host(db, f'{db}', *args))
        htmltexts += data_stats

        self.commit()
        h = render_echart_html('database_cdx', body='<br/>'.join(htmltexts))
        return h

    def dbview_cpu(self, recent=datetime.timedelta(days=1), date_trunc='hour'):
        from pyxllib.data.echarts import render_echart_html

        args = ['CPU核心数（比如4核显示是400%）', date_trunc, recent, 'sum(hosts.cpu_number)*100']

        htmltexts = [
            '<a target="_blank" href="https://www.yuque.com/xlpr/data/hnpb2g?singleDoc#"> 《服务器监控》工具使用文档 </a>']
        res = self._get_host_trace_total('cpu', 'XLPR服务器 CPU 使用近况', *args)
        htmltexts.append(res[0])

        hosts = self.execute('SELECT host_name, nick_name FROM hosts WHERE gpu_gb > 0').fetchall()
        host_stats = []
        for i, (hn, nick_name) in enumerate(hosts, start=1):
            name = f'{hn}，{nick_name}' if nick_name else hn
            host_stats.append(self._get_host_trace_per_host(hn, 'cpu', f'{name}', *args))
        host_stats.sort(key=lambda x: -x[1])  # 按使用量，从多到少排序
        htmltexts += [x[0] for x in host_stats]

        self.commit()

        h = render_echart_html('cpu', body='<br/>'.join(htmltexts))
        return h

    def dbview_cpu_memory(self, recent=datetime.timedelta(days=1), date_trunc='hour'):
        from pyxllib.data.echarts import render_echart_html

        args = ['内存（单位：GB）', date_trunc, recent, 'sum(hosts.cpu_gb)']

        htmltexts = [
            '<a target="_blank" href="https://www.yuque.com/xlpr/data/hnpb2g?singleDoc#"> 《服务器监控》工具使用文档 </a>']
        res = self._get_host_trace_total('cpu_memory', 'XLPR服务器 内存 使用近况', *args)
        htmltexts.append(res[0])

        hosts = self.execute('SELECT host_name, nick_name FROM hosts WHERE gpu_gb > 0').fetchall()
        host_stats = []
        for i, (hn, nick_name) in enumerate(hosts, start=1):
            name = f'{hn}，{nick_name}' if nick_name else hn
            host_stats.append(self._get_host_trace_per_host(hn, 'cpu_memory', f'{name}', *args))
        host_stats.sort(key=lambda x: -x[1])  # 按使用量，从多到少排序
        htmltexts += [x[0] for x in host_stats]

        self.commit()

        h = render_echart_html('cpu_memory', body='<br/>'.join(htmltexts))
        return h

    def dbview_disk_memory(self, recent=datetime.timedelta(days=360), date_trunc='day'):
        """ 查看disk硬盘使用近况
        """
        from pyxllib.data.echarts import render_echart_html

        args = ['硬盘（单位：GB）', date_trunc, recent, 'sum(hosts.disk_gb)']

        htmltexts = [
            '<a target="_blank" href="https://www.yuque.com/xlpr/data/hnpb2g?singleDoc#"> 《服务器监控》工具使用文档 </a>']
        res = self._get_host_trace_total('disk_memory', 'XLPR服务器 DISK硬盘 使用近况', *args)
        htmltexts.append(res[0])
        htmltexts.append('注：xlpr4（四卡）服务器使用du计算/home大小有问题，未统计在列<br/>')

        hosts = self.execute('SELECT host_name, nick_name FROM hosts WHERE gpu_gb > 0').fetchall()
        host_stats = []
        for i, (hn, nick_name) in enumerate(hosts, start=1):
            name = f'{hn}，{nick_name}' if nick_name else hn
            host_stats.append(self._get_host_trace_per_host(hn, 'disk_memory', f'{name}', *args))
        host_stats.sort(key=lambda x: -x[1])  # 按使用量，从多到少排序
        htmltexts += [x[0] for x in host_stats]

        self.commit()

        h = render_echart_html('disk_memory', body='<br/>'.join(htmltexts))
        return h

    def dbview_gpu_memory(self, recent=datetime.timedelta(days=30), date_trunc='day'):
        """ 查看gpu近使用近况

        TODO 这里有可以并行处理的地方，但这个方法不是很重要，不需要特地去做加速占用cpu资源
        """
        from pyxllib.data.echarts import render_echart_html

        args = ['显存（单位：GB）', date_trunc, recent, 'sum(hosts.gpu_gb)']

        htmltexts = [
            '<a target="_blank" href="https://www.yuque.com/xlpr/data/hnpb2g?singleDoc#"> 《服务器监控》工具使用文档 </a>']
        res = self._get_host_trace_total('gpu_memory', 'XLPR八台服务器 GPU显存 使用近况', *args)
        htmltexts.append(res[0])

        hosts = self.execute('SELECT host_name, nick_name FROM hosts WHERE gpu_gb > 0').fetchall()
        host_stats = []
        for i, (hn, nick_name) in enumerate(hosts, start=1):
            name = f'{hn}，{nick_name}' if nick_name else hn
            host_stats.append(self._get_host_trace_per_host(hn, 'gpu_memory', f'{name}', *args))
        host_stats.sort(key=lambda x: -x[1])  # 按使用量，从多到少排序
        htmltexts += [x[0] for x in host_stats]

        self.commit()

        h = render_echart_html('gpu_memory', body='<br/>'.join(htmltexts))
        return h

    def __4_一些数据更新操作(self):
        """ 比如一些扩展字段，在调用api的时候为了性能并没有进行计算，则可以这里补充更新 """

    def update_files_dhash(self, print_mode=True):
        """ 更新files表中的dhash字段值 """
        from pyxllib.cv.imhash import dhash
        from pyxllib.xlcv import xlpil

        # 获取总图片数
        total_count = self.execute("SELECT COUNT(*) FROM files WHERE dhash IS NULL").fetchall()[0][0]

        # 执行查询语句，获取dhash为NULL的记录

        # 初始化进度条
        progress_bar = tqdm(total=total_count, disable=not print_mode)

        while True:
            result = self.execute("SELECT id, data FROM files WHERE dhash IS NULL LIMIT 5")
            if not result:
                break
            for row in result:
                file_id = row[0]
                im = xlpil.read_from_buffer(row[1])
                computed_dhash = str(dhash(im))
                self.update_row('files', {'dhash': computed_dhash}, {'id': file_id})
                progress_bar.update(1)
            self.commit()

    def append_history(self, table_name, where, backup_keys, *,
                       can_merge=None,
                       update_time=None,
                       commit=False):
        """ 为表格添加历史记录，请确保这个表有一个jsonb格式的historys字段

        这里每次都会对关键字段进行全量备份，没有进行高级的优化。
        所以只适用于一些历史记录功能场景。更复杂的还是需要另外自己定制。

        :param table_name: 表名
        :param where: 要记录的id的规则，请确保筛选后记录是唯一的
        :param backup_keys: 需要备份的字段名
        :param can_merge: 在某些情况下，history不需要非常冗余地记录，可以给定与上一条合并的规则
            def can_merge(last, now):
                "last是上一条字典记录，now是当前要记录的字典数据，
                返回True，则用now替换last，并不新增记录"
                ...

        :param update_time: 更新时间，如果不指定则使用当前时间
        """
        # 1 获得历史记录
        ops = ' AND '.join([f'{k}=%s' for k in where.keys()])
        historys = self.exec2one(f'SELECT historys FROM {table_name} WHERE {ops}', list(where.values())) or []
        if historys:
            status1 = historys[-1]
        else:
            status1 = {}

        # 2 获得新记录
        if update_time is None:
            update_time = utc_timestamp()
        status2 = self.exec2dict(f'SELECT {",".join(backup_keys)} FROM {table_name} WHERE {ops}',
                                 list(where.values())).fetchone()
        status2['update_time'] = update_time

        # 3 添加历史记录
        if can_merge is None:
            def can_merge(status1, status2):
                for k in backup_keys:
                    if status1.get(k) != status2.get(k):
                        return False
                return True

        if historys and can_merge(status1, status2):
            historys[-1] = status2
        else:
            historys.append(status2)

        self.update_row(table_name, {'historys': historys}, where, commit=commit)



from arts.openai2._core.chat import Chat, AKPool, system_msg, user_msg, assistant_msg
from arts.openai2._core.group_chat import GroupChat
from arts.openai2._core.chat_in_cmd import chat_in_cmd, ParseCmd